"""

Base implementation for a fasttext tagger; all fasttext taggers should inherit from this class.

@kylel, @soldni

"""
import os
from tempfile import NamedTemporaryFile
from typing import Iterable, Literal, NamedTuple, Optional

import smart_open
from cached_path import cached_path
from fasttext import train_supervised
from fasttext.FastText import _FastText

from .data_types import DocResult, Document, Span, TextSlice
from .taggers import BaseTagger
from .utils import split_paragraphs, split_sentences


class Prediction(NamedTuple):
    label: str
    score: float


class BaseFastTextTagger(BaseTagger):
    SENTENCE_LEVEL_TAGGER = "sentence"
    PARAGRAPH_LEVEL_TAGGER = "paragraph"
    DOCUMENT_LEVEL_TAGGER = "document"

    def __init__(self, model_path: str, model_mode: str) -> None:
        # we use this private attribute to avoid a warning from the fasttext library. See this comment:
        # https://github.com/facebookresearch/fastText/issues/1056#issuecomment-1278058705
        self.classifier = _FastText(str(cached_path(model_path)))
        self.mode = model_mode

    @classmethod
    def train(
        cls,
        train_file: str,
        save_path: str,
        learning_rate: float = 0.1,
        word_vectors_dim: int = 100,
        context_window_size: int = 5,
        max_epochs: int = 100,  # non-default
        min_word_count: int = 1,
        min_label_count: int = 1,
        min_char_ngram: int = 0,
        max_char_ngram: int = 0,
        num_negative_samples: int = 5,
        max_word_ngram: int = 2,  # non-default
        loss_function: Literal["ns", "hs", "softmax", "ova"] = "softmax",
        num_buckets: int = 2_000_000,
        num_threads: int = 0,
        learning_rate_update_rate: int = 100,
        sampling_threshold: float = 0.0001,
        label_prefix: str = "__label__",
        verbose: int = 2,
        pretrained_vectors: Optional[str] = None,
    ) -> _FastText:
        # download potentially remote files
        local_train_file = cached_path(train_file)
        local_pretrained_vectors = cached_path(pretrained_vectors) if pretrained_vectors else None

        # base checks on file format
        with open(local_train_file, "r") as f:
            # check a few lines to see if the file is in the right format
            i = 0
            for ln in f:
                if label_prefix not in ln:
                    raise ValueError(f"{train_file} not the fasttext format, no labels found!")
                if (i := i + 1) > 5:
                    break
            if i == 0:
                raise ValueError(f"{train_file} is empty!")

        # train the fasttext model
        classifier = train_supervised(
            input=local_train_file,
            lr=learning_rate,
            dim=word_vectors_dim,
            ws=context_window_size,
            epoch=max_epochs,
            minCount=min_word_count,
            minCountLabel=min_label_count,
            minn=min_char_ngram,
            maxn=max_char_ngram,
            neg=num_negative_samples,
            wordNgrams=max_word_ngram,
            loss=loss_function,
            bucket=num_buckets,
            thread=num_threads,
            lrUpdateRate=learning_rate_update_rate,
            t=sampling_threshold,
            label=label_prefix,
            verbose=verbose,
            pretrainedVectors=local_pretrained_vectors,
        )

        local_save_path = None
        try:
            # create a local temp file where we save the model
            with NamedTemporaryFile("w", delete=False) as f:
                local_save_path = f.name

            # save the model
            classifier.save_model(local_save_path)

            # upload to remote if save_path is s3 path
            with smart_open.open(save_path, "wb") as fo, smart_open.open(local_save_path, "rb") as fi:
                fo.write(fi.read())
        finally:
            # regardless to what happened, remove the local temp file if it
            # exists
            if local_save_path is not None and os.path.exists(local_save_path):
                os.remove(local_save_path)

        return classifier

    @classmethod
    def test(
        cls,
        test_file: str,
        model_path: Optional[str] = None,
        classifier: Optional[_FastText] = None,
    ):
        # load the model if one is not provided
        if classifier is None:
            assert model_path is not None, "Please provide either a model path or a model"
            classifier = _FastText(str(cached_path(model_path)))

        local_test_file = cached_path(test_file)
        model_performance = classifier.test(local_test_file)
        print(model_performance)

    def predict(self, doc: Document) -> DocResult:
        if self.mode == self.SENTENCE_LEVEL_TAGGER:
            units = split_sentences(doc.text)
        elif self.mode == self.PARAGRAPH_LEVEL_TAGGER:
            units = split_paragraphs(doc.text)
        elif self.mode == self.DOCUMENT_LEVEL_TAGGER:
            units = [TextSlice(doc=doc.text, start=0, end=len(doc.text))]
        else:
            raise ValueError(f"Unknown mode {self.mode}")

        spans = []
        for unit in units:
            for prediction in self.predict_slice(unit):
                spans.append(Span(start=unit.start, end=unit.end, type=prediction.label, score=prediction.score))

        return DocResult(doc=doc, spans=spans)

    def predict_slice(self, text_slice: TextSlice) -> Iterable[Prediction]:
        raise NotImplementedError("Please implement the predict slice method")



"""

Data types assumed by Filters.

@kylel, @soldni

"""

from typing import Any, Dict, List, Optional, Tuple, Union

from msgspec import Struct


class InputSpec(Struct):
    id: str
    text: str
    source: str
    version: Optional[str] = None


class OutputSpec(Struct):
    id: str
    attributes: Dict[str, List[List[Union[int, float]]]]
    source: Optional[str] = None


class Document:
    __slots__ = "source", "version", "id", "text"

    def __init__(self, source: str, id: str, text: str, version: Optional[str] = None) -> None:
        self.source = source
        self.version = version
        self.id = id
        self.text = text

    @classmethod
    def from_spec(cls, spec: InputSpec) -> "Document":
        return Document(source=spec.source, version=spec.version, id=spec.id, text=spec.text)

    def to_spec(self) -> InputSpec:
        return InputSpec(source=self.source, version=self.version, id=self.id, text=self.text)

    @classmethod
    def from_json(cls, d: Dict) -> "Document":
        return Document(source=d["source"], version=d["version"], id=d["id"], text=d["text"])

    def to_json(self) -> Dict:
        return {"source": self.source, "version": self.version, "id": self.id, "text": self.text}

    def __str__(self) -> str:
        return (
            str(self.__class__.__name__)
            + f"(source={repr(self.source)},version={repr(self.version)},id={repr(self.id)},text={repr(self.text)})"
        )


class Span:
    __slots__ = "start", "end", "type", "score", "experiment", "tagger"

    def __init__(
        self,
        start: int,
        end: int,
        type: str,
        score: float = 1.0,
        experiment: Optional[str] = None,
        tagger: Optional[str] = None,
    ):
        self.start = start
        self.end = end
        self.type = type
        self.score = float(score)
        self.experiment = experiment
        self.tagger = tagger

    def mention(self, text: str, window: int = 0) -> str:
        return text[max(0, self.start - window) : min(len(text), self.end + window)]

    def select(self, doc: Document) -> str:
        return doc.text[self.start : self.end]

    @classmethod
    def from_spec(cls, attribute_name: str, attribute_value: List[Union[int, float]]) -> "Span":
        if "__" in attribute_name:
            # bff tagger has different name
            exp_name, tgr_name, attr_type = attribute_name.split("__", 2)
        else:
            exp_name = tgr_name = attr_type = attribute_name

        start, end, score = attribute_value
        return Span(
            start=int(start),
            end=int(end),
            type=attr_type,
            score=float(score),
            experiment=exp_name,
            tagger=tgr_name,
        )

    def to_spec(self) -> Tuple[str, List[Union[int, float]]]:
        assert self.experiment is not None, "Experiment name must be set to convert to spec"
        assert self.tagger is not None, "Tagger name must be set to convert to spec"
        return (
            f"{self.experiment}__{self.tagger}__{self.type}",
            [self.start, self.end, self.score],
        )

    def __len__(self) -> int:
        return self.end - self.start

    @classmethod
    def from_json(cls, di: Dict) -> "Span":
        return Span(start=di["start"], end=di["end"], type=di["type"], score=di["score"])

    def to_json(self, text: Optional[str] = None, window: int = 0) -> dict:
        span_repr = {"start": self.start, "end": self.end, "type": self.type, "score": self.score}
        if text is not None:
            span_repr["mention"] = self.mention(text=text, window=window)
        return span_repr

    def __str__(self) -> str:
        cls_name = self.__class__.__name__
        return f"{cls_name}(start={self.start},end={self.end},type={repr(self.type)},score={self.score:.5f})"


class DocResult:
    __slots__ = "doc", "spans"

    def __init__(self, doc: Document, spans: List[Span]) -> None:
        self.doc = doc
        self.spans = spans

    @classmethod
    def from_spec(cls, doc: InputSpec, *attrs_groups: OutputSpec) -> "DocResult":
        spans: List[Span] = []
        for attrs in attrs_groups:
            assert doc.id == attrs.id, f"doc.id={doc.id} != attrs.id={attrs.id}"
            spans.extend(
                [
                    Span.from_spec(attribute_name=attr_name, attribute_value=attr_value)
                    for attr_name, attr_values in attrs.attributes.items()
                    for attr_value in attr_values
                ]
            )
        return DocResult(doc=Document.from_spec(doc), spans=spans)

    def to_spec(self) -> Tuple[InputSpec, OutputSpec]:
        doc_spec = self.doc.to_spec()
        attributes: Dict[str, List[List[Union[int, float]]]] = {}

        for span in self.spans:
            attr_name, attr_value = span.to_spec()
            attributes.setdefault(attr_name, []).append(attr_value)

        return doc_spec, OutputSpec(source=self.doc.source, id=self.doc.id, attributes=attributes)

    @classmethod
    def from_json(cls, d: Dict[str, Any]) -> "DocResult":
        return DocResult(
            doc=Document.from_json(d["doc"]),
            spans=[Span.from_json(span) for span in d["spans"]],
        )

    def to_json(self, with_doc: bool = False, window: int = 0) -> Dict[str, Any]:
        d: Dict[str, Any] = {"spans": [span.to_json(text=self.doc.text, window=window) for span in self.spans]}
        if with_doc:
            d["doc"] = self.doc.to_json()
        return d

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(doc={self.doc},spans=[{','.join(str(s) for s in self.spans)}])"


class TextSlice:
    """A slice of text from a document."""

    __slots__ = "doc", "start", "end"

    def __init__(self, doc: str, start: int, end: int):
        self.doc = doc
        self.start = start
        self.end = end

    @property
    def text(self) -> str:
        return self.doc[self.start : self.end]

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(text={repr(self.text)},start={self.start},end={self.end})"



from pathlib import Path
from unittest import TestCase

from dolma.core.runtime import _make_paths_from_prefix, _make_paths_from_substitution

LOCAL_DATA = Path(__file__).parent.parent / "data"


class TestRuntimeUtilities(TestCase):
    def test_make_paths_from_substitution(self):
        paths = [
            "s3://bucket/common-crawl/documents/cc_*/*.json.gz",
            "/local/path/to/documents/train/*",
        ]
        new_paths = _make_paths_from_substitution(
            paths=paths,
            find="documents",
            replace="attributes",
        )
        self.assertEqual(new_paths, ["s3://bucket/common-crawl/attributes", "/local/path/to/attributes/train"])

    def test_make_paths_from_prefix(self):
        paths = [
            "s3://bucket/common-crawl/documents/cc_head/*.json.gz",
            "s3://bucket/common-crawl/documents/cc_middle/*.json.gz",
            "s3://bucket/common-crawl/documents/cc_tail/*.json.gz",
        ]
        new_paths = _make_paths_from_prefix(
            paths=paths,
            prefix="s3://bucket/common-crawl/attributes/",
        )
        self.assertEqual(
            new_paths,
            [
                "s3://bucket/common-crawl/attributes/cc_head",
                "s3://bucket/common-crawl/attributes/cc_middle",
                "s3://bucket/common-crawl/attributes/cc_tail",
            ],
        )

        paths = [
            "s3://bucket/common-crawl/documents/*.json.gz",
            "s3://bucket2/c4/documents/**/data/*.json.gz",
        ]
        new_paths = _make_paths_from_prefix(
            paths=paths,
            prefix="/local/path/",
        )
        self.assertEqual(
            new_paths,
            [
                "/local/path/bucket/common-crawl/documents",
                "/local/path/bucket2/c4/documents",
            ],
        )



"""
Utilities to work with a OmegaConf structured config object

Author: Luca Soldaini (@soldni)
"""


from argparse import ArgumentParser, Namespace
from collections.abc import Iterable
from copy import deepcopy
from dataclasses import Field
from dataclasses import field as dataclass_field
from dataclasses import is_dataclass
from logging import warn
from typing import Any, Dict, Generic, Literal, Optional, Protocol, Type, TypeVar, Union

from omegaconf import MISSING, DictConfig, ListConfig
from omegaconf import OmegaConf as om
from rich.console import Console
from rich.syntax import Syntax

__all__ = [
    "BaseCli",
    "field",
    "make_parser",
    "namespace_to_nested_omegaconf",
    "print_config",
]


T = TypeVar("T", bound=Any)
D = TypeVar("D", bound="DataClass")
A = TypeVar("A", bound="ArgumentParser")


def _field_nargs(default: Any) -> Union[Literal["?"], Literal["*"]]:
    # return '+' if _default is iterable but not string/bytes, else 1
    if isinstance(default, str) or isinstance(default, bytes):
        return "?"
    elif isinstance(default, Iterable):
        return "*"
    else:
        return "?"


def field(default: T = MISSING, help: Optional[str] = None, **extra: Any) -> T:
    metadata = {"help": help, "type": type(default), "default": default, "nargs": _field_nargs(default), **extra}
    return dataclass_field(default_factory=lambda: deepcopy(default), metadata=metadata)


class DataClass(Protocol):
    __dataclass_fields__: Dict[str, Field]


def make_parser(parser: A, config: Type[DataClass], prefix: Optional[str] = None) -> A:
    for field_name, field in config.__dataclass_fields__.items():
        # get type from annotations or metadata
        typ_ = config.__annotations__.get(field_name, field.metadata.get("type", MISSING))

        if typ_ is MISSING:
            warn(f"No type annotation for field {field_name} in {config.__name__}")
            continue

        if is_dataclass(typ_):
            # recursively add subparsers
            make_parser(parser, typ_, prefix=field_name)
            continue

        field_name = f"{prefix}.{field_name}" if prefix else field_name
        parser.add_argument(
            f"--{field_name}",
            help=field.metadata.get("help"),
            nargs=field.metadata.get("nargs", "?"),
            default=MISSING,
        )

    return parser


def _make_nested_dict(key: str, value: Any, d: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    d = d or {}

    if "." in key:
        key, rest = key.split(".", 1)
        value = _make_nested_dict(rest, value, d.get(key))

    if value is not MISSING:
        d[key] = value

    return d


def namespace_to_nested_omegaconf(args: Namespace, structured: Type[T], config: Optional[dict] = None) -> T:
    nested_config_dict: Dict[str, Any] = {}
    for key, value in vars(args).items():
        nested_config_dict = _make_nested_dict(key, value, nested_config_dict)

    untyped_config: DictConfig = om.merge(
        om.create(config or {}), om.create(nested_config_dict)
    )  # pyright: ignore (pylance is confused because om.create might return a DictConfig or a ListConfig)

    base_structured_config: DictConfig = om.structured(structured)
    merged_config = om.merge(base_structured_config, untyped_config)
    assert isinstance(merged_config, DictConfig)
    return merged_config  # pyright: ignore


def print_config(config: Any, console: Optional[Console] = None) -> None:
    if not isinstance(config, (DictConfig, ListConfig)):
        config = om.create(config)
    console = console or Console()
    syntax = Syntax(code=om.to_yaml(config).strip(), lexer="yaml", theme="ansi_dark")
    console.print(syntax)


class BaseCli(Generic[D]):
    CONFIG: Type[D]

    @classmethod
    def make_parser(cls, parser: A) -> A:
        assert hasattr(cls, "CONFIG"), f"{cls.__name__} must have a CONFIG attribute"
        return make_parser(parser, cls.CONFIG)

    @classmethod
    def run_from_args(cls, args: Namespace, config: Optional[dict] = None):
        assert hasattr(cls, "CONFIG"), f"{cls.__name__} must have a CONFIG attribute"
        parsed_config = namespace_to_nested_omegaconf(args=args, structured=cls.CONFIG, config=config)
        return cls.run(parsed_config)

    @classmethod
    def run(cls, parsed_config: D):
        raise NotImplementedError("Abstract method; must be implemented in subclass")



import logging
from collections import Counter
from dataclasses import dataclass
from statistics import median
from typing import Counter as CounterType
from typing import List, Tuple

from ..core.data_types import DocResult, Document, Span
from ..core.registry import TaggerRegistry
from ..core.taggers import BaseTagger

REQUIRED_ENGLISH_WORDS = {"the", "be", "to", "of", "and", "that", "have", "with"}
SYMBOLS = {"#", "\u2026"}
BULLET_POINTS = {"*", "-"}


@dataclass
class GopherAttributes:
    fraction_of_characters_in_most_common_ngram: List[Tuple[int, float]]
    fraction_of_characters_in_duplicate_ngrams: List[Tuple[int, float]]
    character_count: int = 0
    word_count: int = 0
    median_word_length: float = False
    symbol_to_word_ratio: float = 0.0
    fraction_of_words_with_alpha_character: float = 0.0
    required_word_count: int = 0
    fraction_of_lines_starting_with_bullet_point: float = 0.0
    fraction_of_lines_ending_with_ellipsis: float = 0.0
    fraction_of_duplicate_lines: float = 0.0
    fraction_of_characters_in_duplicate_lines: float = 0.0

    def as_spans(self) -> List[Span]:
        spans = []
        spans.extend(
            [
                Span(0, self.character_count, f"fraction_of_characters_in_most_common_{n}grams", v)
                for n, v in self.fraction_of_characters_in_most_common_ngram
            ]
        )
        spans.extend(
            [
                Span(0, self.character_count, f"fraction_of_characters_in_duplicate_{n}grams", v)
                for n, v in self.fraction_of_characters_in_duplicate_ngrams
            ]
        )
        spans.append(Span(0, self.character_count, type="character_count", score=self.character_count))
        spans.append(Span(0, self.character_count, type="word_count", score=self.word_count))
        spans.append(Span(0, self.character_count, type="median_word_length", score=self.median_word_length))
        spans.append(Span(0, self.character_count, type="symbol_to_word_ratio", score=self.symbol_to_word_ratio))
        spans.append(
            Span(
                0,
                self.character_count,
                type="fraction_of_words_with_alpha_character",
                score=self.fraction_of_words_with_alpha_character,
            )
        )
        spans.append(Span(0, self.character_count, type="required_word_count", score=self.required_word_count))
        spans.append(
            Span(
                0,
                self.character_count,
                type="fraction_of_lines_starting_with_bullet_point",
                score=self.fraction_of_lines_starting_with_bullet_point,
            )
        )
        spans.append(
            Span(
                0,
                self.character_count,
                type="fraction_of_lines_ending_with_ellipsis",
                score=self.fraction_of_lines_ending_with_ellipsis,
            )
        )
        spans.append(
            Span(
                0, self.character_count, type="fraction_of_duplicate_lines", score=self.fraction_of_duplicate_lines
            )
        )
        spans.append(
            Span(
                0,
                self.character_count,
                type="fraction_of_characters_in_duplicate_lines",
                score=self.fraction_of_characters_in_duplicate_lines,
            )
        )
        return spans


def get_attributes(text: str) -> GopherAttributes:
    attrs = GopherAttributes([], [])
    attrs.character_count = len(text)
    if attrs.character_count == 0:
        return attrs

    try:
        words = text.split()
        word_count = len(words)
        character_count = sum(len(word) for word in words)

        attrs.word_count = word_count
        attrs.median_word_length = median([len(word) for word in words])
        attrs.symbol_to_word_ratio = sum(1 for word in words if any(s in word for s in SYMBOLS)) / word_count
        attrs.fraction_of_words_with_alpha_character = (
            sum(1 for word in words if any(c.isalpha() for c in word)) / word_count
        )
        attrs.required_word_count = sum(1 for word in words if word in REQUIRED_ENGLISH_WORDS)

        all_counts = all_ngram_counts(words)

        count_most_common_ngrams = {2, 3, 4}
        for n, ngram_counts in all_counts:
            if not ngram_counts:
                continue
            if n in count_most_common_ngrams:
                most_common_ngram, count = ngram_counts.most_common(1)[0]
                value = count * sum(len(w) for w in most_common_ngram) / character_count
                attrs.fraction_of_characters_in_most_common_ngram.append((n, value))
            else:
                ng_char_count = sum(count * sum(len(w) for w in ng) for ng, count in ngram_counts.items())
                value = (
                    sum(count * sum(len(w) for w in ng) for ng, count in ngram_counts.items() if count > 1)
                    / ng_char_count
                )
                attrs.fraction_of_characters_in_duplicate_ngrams.append((n, value))

        lines = text.split("\n")
        line_count = len(lines)
        for line in lines:
            if any(line.startswith(s) for s in BULLET_POINTS):
                attrs.fraction_of_lines_starting_with_bullet_point += 1
            if line.endswith("\u2026"):
                attrs.fraction_of_lines_ending_with_ellipsis += 1
        attrs.fraction_of_lines_starting_with_bullet_point /= line_count
        attrs.fraction_of_lines_ending_with_ellipsis /= line_count

        line_counts = Counter(lines)
        attrs.fraction_of_duplicate_lines = (
            sum(count for line, count in line_counts.items() if count > 1) / line_count
        )
        attrs.fraction_of_characters_in_duplicate_lines = (
            sum(len(line) * count for line, count in line_counts.items() if count > 1) / character_count
        )
    except Exception as e:
        logging.exception(f"Error processing text {e}: {text[:200]}")

    return attrs


def all_ngram_counts(words) -> List[Tuple[int, CounterType[Tuple[str, ...]]]]:
    return [(n, Counter(list(zip(*[words[i:] for i in range(n)])))) for n in range(2, 11)]


def all_ngram_counts_alt(words: List[str]) -> List[Tuple[int, CounterType[Tuple[str, ...]]]]:
    """Seems like it should be faster, but isn't"""
    ngram: List[Tuple[str, ...]] = list(zip(words, words[1:]))
    all_counts: List[Tuple[int, CounterType[Tuple[str, ...]]]] = [(2, Counter(ngram))]

    for n in range(3, 11):
        ngram = list(a + (b,) for a, b in zip(ngram, words[n - 1 :]))
        all_counts.append((n, Counter(ngram)))

    return all_counts


@TaggerRegistry.add("gopher_v1")
class GopherTagger(BaseTagger):
    def predict(self, doc: Document) -> DocResult:
        attrs = get_attributes(doc.text)
        result = DocResult(doc=doc, spans=attrs.as_spans())
        return result



import json
import warnings

# warning raised by pkg_resources used in a lot of google packages
warnings.filterwarnings("ignore", message=r".*declare_namespace\(\'.*google.*", category=DeprecationWarning)
# base warning raised when warning above are raised
warnings.filterwarnings("ignore", message=r".*pkg_resources is deprecated.*", category=DeprecationWarning)

# must import taggers to register them
# we import the rust extension here and wrap it in a python module
from . import dolma as _dolma  # type: ignore   # noqa: E402
from .core.errors import DolmaRustPipelineError  # noqa: E402
from .taggers import *  # noqa: E402


def deduper(config: dict):
    try:
        _dolma.deduper_entrypoint(json.dumps(config))
    except RuntimeError as e:
        raise DolmaRustPipelineError(f"Error running deduper: {e}") from e


def mixer(config: dict):
    try:
        _dolma.mixer_entrypoint(json.dumps(config))
    except RuntimeError as e:
        raise DolmaRustPipelineError(f"Error running mixer: {e}") from e



from dataclasses import dataclass
from pathlib import Path
from tempfile import gettempdir

from dolma.cli import field


@dataclass
class WorkDirConfig:
    input: str = field(
        default=str(Path(gettempdir()) / "dolma" / "deduper" / "input"),
        help="Path to the input directory. Required.",
    )
    output: str = field(
        default=str(Path(gettempdir()) / "dolma" / "deduper" / "output"),
        help="Path to the output directory. Required.",
    )



"""

Unit tests for core/data_types.py

@kylel

"""

from unittest import TestCase

from dolma.core.data_types import DocResult, Document, InputSpec, Span, TextSlice


class TestDocument(TestCase):
    def test_document_to_from_json(self):
        doc = Document(source="source", version="version", id="id", text="text")
        doc_json = doc.to_json()
        doc_json2 = {
            "source": "source",
            "version": "version",
            "id": "id",
            "text": "text",
        }
        self.assertEqual(doc_json, doc_json2)
        doc2 = Document.from_json(doc_json2)
        self.assertEqual(doc_json, doc2.to_json())

    def test_document_to_from_spec(self):
        doc = Document(source="source", version="version", id="id", text="text")
        spec = doc.to_spec()
        spec2 = InputSpec(source="source", version="version", id="id", text="text")
        self.assertEqual(spec, spec2)
        doc2 = Document.from_spec(spec2)
        self.assertEqual(spec, doc2.to_spec())


class TestSpan(TestCase):
    def test_span_to_from_json(self):
        span = Span(start=0, end=1, type="type", score=1.0)
        span_json = span.to_json()
        span_json2 = {"start": 0, "end": 1, "type": "type", "score": 1.0}
        self.assertEqual(span_json, span_json2)
        span2 = Span.from_json(span_json2)
        self.assertEqual(span_json, span2.to_json())

    # TODO: add tests for to/from Spec
    def test_span_to_from_spec(self):
        span = Span(start=0, end=1, type="type", score=1.0)
        with self.assertRaises(AssertionError):
            span.to_spec()


class TestDocResult(TestCase):
    def test_doc_result_to_from_json(self):
        doc = Document(source="source", version="version", id="id", text="text")
        spans = [
            Span(start=0, end=2, type="xxx", score=1.0),
            Span(start=2, end=4, type="yyy", score=0.5),
        ]
        doc_result = DocResult(doc=doc, spans=spans)

        # to_json() doesnt return Document by default
        # also, it returns this extra field called `"mention"`
        doc_result_json = doc_result.to_json()
        doc_result_json2 = {
            "spans": [
                {"start": 0, "end": 2, "type": "xxx", "score": 1.0, "mention": "te"},
                {"start": 2, "end": 4, "type": "yyy", "score": 0.5, "mention": "xt"},
            ]
        }
        self.assertEqual(doc_result_json, doc_result_json2)

        # from_json() requires also providing the Document
        with self.assertRaises(KeyError):
            DocResult.from_json(doc_result_json2)
        doc_result_json3 = {
            "doc": {
                "source": "source",
                "version": "version",
                "id": "id",
                "text": "text",
            },
            "spans": [
                {"start": 0, "end": 2, "type": "xxx", "score": 1.0, "mention": "te"},
                {"start": 2, "end": 4, "type": "yyy", "score": 0.5, "mention": "xt"},
            ],
        }
        doc_result3 = DocResult.from_json(doc_result_json3)
        self.assertEqual(doc_result_json, doc_result3.to_json())


class TestTextSlice(TestCase):
    def test_text_slice_text(self):
        text = "This is a test"
        slice = TextSlice(doc=text, start=0, end=4)
        self.assertEqual(slice.text, "This")



from .data_types import DocResult, Document, Span
from .registry import TaggerRegistry
from .taggers import BaseTagger

__all__ = [
    "BaseTagger",
    "DocResult",
    "Document",
    "Span",
    "TaggerRegistry",
]



import json
import os
import re
import uuid
from typing import List, Tuple
from urllib.parse import urlparse

import boto3
import smart_open
from smart_open import open

from dolma.core.paths import glob_path, mkdir_p

DOLMA_TEST_S3_PREFIX_ENV_VAR = "DOLMA_TESTS_S3_PREFIX"
DOLMA_TEST_S3_PREFIX_DEFAULT = "s3://dolma-tests"


def parse_s3_path(s3_path: str) -> Tuple[str, str]:
    """
    Parse an s3 path into a bucket and key.

    Args:
        s3_path: The s3 path to parse.

    Returns:
        A tuple containing the bucket and key.
    """
    if not re.match(r"^s3://[\w-]+", s3_path):
        raise RuntimeError(f"Invalid s3 path: {s3_path}")

    # use urlparse to parse the s3 path
    parsed = urlparse(s3_path)
    return parsed.netloc, parsed.path.lstrip("/")


def get_test_prefix() -> str:
    # get the test prefix from the environment, or use the default if not set
    test_prefix = os.environ.get(DOLMA_TEST_S3_PREFIX_ENV_VAR, DOLMA_TEST_S3_PREFIX_DEFAULT)

    # this will check if it is a valid path
    bucket, _ = parse_s3_path(test_prefix)

    # check if the user has access to the test bucket using boto3
    s3 = boto3.client("s3")
    try:
        s3.head_bucket(Bucket=bucket)
    except Exception:
        raise RuntimeError(
            f"Unable to access test bucket '{test_prefix}'. To provide a different bucket, "
            f"set the '{DOLMA_TEST_S3_PREFIX_ENV_VAR}' environment variable before running the tests."
        )

    # add a uuid to the test prefix to avoid collisions
    return f"{test_prefix.rstrip()}/{uuid.uuid4()}"


def upload_test_documents(local_input: str, test_prefix: str) -> Tuple[str, str]:
    remote_input = f"{test_prefix}/input/documents"
    remote_output = f"{test_prefix}/output/documents"

    for i, local_fp in enumerate(glob_path(local_input)):
        remote_fp = f"{remote_input}/{i:05d}.json.gz"

        with open(local_fp, "rb") as f, open(remote_fp, "wb") as g:
            g.write(f.read())

    return remote_input, remote_output


def upload_test_attributes(local_attributes: str, test_prefix: str):
    remote_attributes = f"{test_prefix}/input/attributes"

    for i, local_fp in enumerate(glob_path(local_attributes)):
        matched = re.match(r"^(attributes|duplicate)-(\w+)", local_fp)
        if not matched:
            raise RuntimeError(f"Unexpected filename: {local_fp}")

        _, name = matched.groups()

        remote_fp = f"{remote_attributes}/{name}/{i:05d}.json.gz"

        with open(local_fp, "rb") as f, open(remote_fp, "wb") as g:
            g.write(f.read())


def clean_test_data(test_prefix: str):
    s3 = boto3.client("s3")

    bucket_name, prefix = parse_s3_path(test_prefix)

    for obj in s3.list_objects_v2(Bucket=bucket_name, Prefix=prefix).get("Contents", []):
        s3.delete_object(Bucket=bucket_name, Key=obj["Key"])


def download_s3_prefix(s3_prefix: str, local_prefix: str):
    s3 = boto3.client("s3")

    bucket_name, prefix = parse_s3_path(s3_prefix)

    for obj in s3.list_objects_v2(Bucket=bucket_name, Prefix=prefix).get("Contents", []):
        name = obj["Key"].replace(prefix, "").lstrip("/")
        local_fp = os.path.join(local_prefix, name)
        mkdir_p(os.path.dirname(local_fp))

        s3.download_file(Bucket=bucket_name, Key=obj["Key"], Filename=local_fp)


def upload_s3_prefix(s3_prefix: str, local_prefix: str):
    s3 = boto3.client("s3")

    bucket_name, prefix = parse_s3_path(s3_prefix)

    for local_fp in glob_path(local_prefix):
        name = local_fp.replace(local_prefix, "").lstrip("/")
        s3.upload_file(Bucket=bucket_name, Key=f"{prefix}/{name}", Filename=local_fp)


def load_jsonl(fp: str) -> List[dict]:
    with smart_open.open(fp, "r") as f:
        return [json.loads(ln) for ln in f]



import inspect
import itertools
import multiprocessing
import pickle
import random
import re
import time
from contextlib import ExitStack
from datetime import datetime
from functools import partial
from queue import Queue
from threading import Thread
from typing import Any, Dict, List, Optional, Tuple, Union

import smart_open
import tqdm

from .errors import DolmaError, DolmaRetryableFailure
from .paths import add_suffix, glob_path, make_relative, mkdir_p, sub_prefix

METADATA_SUFFIX = ".done.txt"


class BaseParallelProcessor:
    """A base parallel processor that supports applying the same process_single method to a list of files.

    This class is meant to be subclassed. The subclass must implement:
        - `process_single` method, which takes a source path file to transform, and a destination path where
           to save the transformed file.
        - `increment_progressbar` method, which defines which units to keep track of in the progress bar.

    See documentation of both methods for more details on how to implement them correctly.
    """

    def __init__(
        self,
        source_prefix: Union[str, List[str]],
        destination_prefix: Union[str, List[str]],
        metadata_prefix: Union[str, List[str]],
        num_processes: int = 1,
        debug: bool = False,
        seed: int = 0,
        pbar_timeout: float = 1e-3,
        ignore_existing: bool = False,
        include_paths: Optional[List[str]] = None,
        exclude_paths: Optional[List[str]] = None,
        files_regex_pattern: Optional[str] = None,
        retries_on_error: int = 0,
    ):
        """Initialize the parallel processor.

        Args:
            source_prefix (str): The location where source files are stored. This can be a local directory or a
                prefix to an S3 location.
            destination_prefix (str): The location where to save the transformed files. This can be a local
                directory or a prefix to an S3 location. Local directories will be created if they do not exist.
                The directory structure from the source prefix will be replicated in the destination prefix;
                file names will also be the same.
            metadata_prefix (str): The prefix of the metadata files to save. This can be a local path or an
                S3 path. Metadata output will be created for each file after it is processed. Filenames are
                checked to verify if a file has been processed and can be skipped unless `ignore_existing` is
                set to true.
            num_processes (int, optional): The number of processes to use. Defaults to 1.
            debug (bool, optional): Whether to run in debug mode; if true, no multiprocessing will be used.
                Defaults to False.
            seed (int, optional): The random seed to use when shuffling input files. Defaults to 0.
            pbar_timeout (float, optional): How often to update progress bars in seconds.
                Defaults to 0.01 seconds.
            ignore_existing (bool, optional): Whether to ignore files that have been already processed and
                re-run the processor on all files from scratch. Defaults to False.
            include_paths (Optional[List[str]], optional): A list of paths to include. If provided, only files
                that match one of the paths will be processed. Defaults to None.
            exclude_paths (Optional[List[str]], optional): A list of paths to exclude. If provided, files that
                match one of the paths will be skipped. Defaults to None.
        """

        self.src_prefixes = [source_prefix] if isinstance(source_prefix, str) else source_prefix
        self.dst_prefixes = [destination_prefix] if isinstance(destination_prefix, str) else destination_prefix
        self.meta_prefixes = [metadata_prefix] if isinstance(metadata_prefix, str) else metadata_prefix
        self.num_processes = num_processes
        self.debug = debug
        self.seed = seed
        self.pbar_timeout = pbar_timeout
        self.ignore_existing = ignore_existing

        self.include_paths = set(include_paths) if include_paths is not None else None
        self.exclude_paths = set(exclude_paths) if exclude_paths is not None else None
        self.files_regex_pattern = re.compile(files_regex_pattern) if files_regex_pattern else None
        self.retries_on_error = retries_on_error

        # checking that the increment_progressbar method is subclassed correctly
        sig = inspect.signature(self.increment_progressbar)
        if "queue" not in sig.parameters or sig.parameters["queue"].kind != inspect.Parameter.POSITIONAL_ONLY:
            raise AttributeError(
                "increment_progressbar must have a positional-only argument named 'queue'; "
                "Check that you have subclassed BaseParallelProcessor correctly!"
            )
        if "kwargs" in sig.parameters and sig.parameters["kwargs"].kind == inspect.Parameter.VAR_KEYWORD:
            raise AttributeError(
                "increment_progressbar must not have a **kwargs argument; "
                "Check that you have subclassed BaseParallelProcessor correctly!"
            )
        if any(p.name != "queue" and p.default != 0 for p in sig.parameters.values()):
            raise AttributeError(
                "increment_progressbar must have a default value of 0 for all arguments except 'queue'; "
                "Check that you have subclassed BaseParallelProcessor correctly!"
            )

        if len(self.src_prefixes) != len(self.dst_prefixes) or len(self.src_prefixes) != len(self.meta_prefixes):
            raise ValueError("The number of source, destination and metadata prefixes must be the same.")

        if len(self.src_prefixes) == 0:
            raise ValueError("At least one source prefix must be provided.")

        if any("*" in p for p in itertools.chain(self.dst_prefixes, self.meta_prefixes)):
            raise ValueError("Destination and metadata prefixes cannot contain wildcards.")

        for i in range(len(self.src_prefixes)):
            # adding a wildcard to the end of the each source prefix if it doesn't have one
            self.src_prefixes[i] = add_suffix(p, "*") if "*" not in (p := self.src_prefixes[i]) else p

    @classmethod
    def process_single(
        cls,
        source_path: str,
        destination_path: str,
        queue: "Queue[Union[None, Tuple[int, ...]]]",
        **kwargs: Any,
    ):
        """Process a single file.

        This method must be implemented by the subclass. It takes a source path file to transform, and a
        destination path where to save the transformed file. It also takes a queue to increment the progress
        bars. The queue should be passed to the `increment_progressbar` method.

        Args:
            source_path (str): The path to the source file to transform. Can be an S3 path or a local path.
            destination_path (str): The path to the destination file to save. Can be an S3 path or a local path.
            queue (Queue[Union[None, Tuple[int, ...]]]): The queue to increment the progress bars.
        """
        raise NotImplementedError()

    @classmethod
    def _process_single_and_save_status(
        cls,
        source_path: str,
        destination_path: str,
        metadata_path: str,
        queue: "Queue[Union[None, Tuple[int, ...]]]",
        serialized_kwargs: bytes,
    ):
        """A wrapper around process single that saves a metadata file if processing is successful."""

        kwargs = pickle.loads(serialized_kwargs)
        retries_on_error = kwargs.get("retries_on_error", 0) + 1
        while True:
            try:
                cls.process_single(
                    source_path=source_path, destination_path=destination_path, queue=queue, **kwargs
                )
                break
            except DolmaRetryableFailure as e:
                retries_on_error -= 1
                if retries_on_error == 0:
                    raise DolmaError from e

        with smart_open.open(metadata_path, "wt") as f:
            f.write(datetime.now().isoformat())

    @classmethod
    def increment_progressbar(
        cls, queue: "Queue[Union[None, Tuple[int, ...]]]", /, **kwargs: int
    ) -> Dict[str, int]:
        """Increment the progress bar by putting a tuple in the queue.

        When subclassing, we recommend defining which units to keep track of in the progress bar by
        defining keyword arguments. Then you can call the base class via `super()` and pass the keyword.
        Example:

        ```python
        class MyProcessor(BaseParallelProcessor):
            def increment_progressbar(self, queue, /, files = 0, documents = 0):   # we use two progress bars
                return super().increment_progressbar(queue, files=files, documents=documents)
        ```
        """
        queue.put(tuple(kwargs.get(k, 0) for k in kwargs))
        return kwargs

    @classmethod
    def _run_threaded_progressbar(
        cls,
        queue: "Queue[Union[None, Tuple[int, ...]]]",
        timeout: float,
    ):
        """Run a progress bar in a separate thread.

        Args:
            queue (Queue[Union[None, Tuple[int, ...]]]): The queue to increment the progress bars.
            timeout (float): How often to update the progress bars in seconds.
        """

        sample_queue_output = cls.increment_progressbar(queue)

        with ExitStack() as stack:
            pbars = [
                stack.enter_context(
                    tqdm.tqdm(desc=str(k), unit=str(k)[:1], position=i, unit_scale=True)  # pyright: ignore
                )
                for i, k in enumerate(sample_queue_output)
            ]

            while True:
                item = queue.get()
                if item is None:
                    break

                for pbar, value in zip(pbars, item):
                    pbar.update(value)

                time.sleep(timeout)

    def _debug_run_all(
        self,
        all_source_paths: List[str],
        all_destination_paths: List[str],
        all_metadata_paths: List[str],
        **process_single_kwargs: Any,
    ):
        """Run files one by one on the main process

        Args:
            all_source_paths (List[MultiPath]): The list of source paths to process.
            all_destination_paths (List[MultiPath]): The list of destination paths to save.
            all_metadata_paths (List[MultiPath]): The locations where to save metadata.
        """

        it = zip(all_source_paths, all_destination_paths, all_metadata_paths)
        pbar_queue: "Queue[Union[None, Tuple[int, ...]]]" = Queue()
        thread = Thread(target=self._run_threaded_progressbar, args=(pbar_queue, self.pbar_timeout), daemon=True)
        thread.start()

        for source_prefix, destination_prefix, metadata_prefix in it:
            self._process_single_and_save_status(
                source_path=source_prefix,
                destination_path=destination_prefix,
                metadata_path=metadata_prefix,
                queue=pbar_queue,
                serialized_kwargs=pickle.dumps(process_single_kwargs),
            )

        pbar_queue.put(None)
        thread.join()

    def _multiprocessing_run_all(
        self,
        all_source_paths: List[str],
        all_destination_paths: List[str],
        all_metadata_paths: List[str],
        **process_single_kwargs: Any,
    ):
        """Run files in parallel using multiprocessing.

        Args:
            all_source_paths (List[MultiPath]): The list of source paths to process.
            all_destination_paths (List[MultiPath]): The list of destination paths to save.
            all_metadata_paths (List[MultiPath]): The locations where to save metadata.
        """
        try:
            multiprocessing.set_start_method("spawn")
        except RuntimeError:
            assert multiprocessing.get_start_method() == "spawn", "Multiprocessing start method must be spawn"

        with multiprocessing.Pool(processes=self.num_processes) as pool:
            pbar_queue: "Queue[Union[None, Tuple[int, ...]]]" = (manager := multiprocessing.Manager()).Queue()
            thread = Thread(
                target=self._run_threaded_progressbar, args=(pbar_queue, self.pbar_timeout), daemon=True
            )
            thread.start()

            process_single_fn = partial(self.process_single, queue=pbar_queue)
            results = []

            for s, d, m in zip(all_source_paths, all_destination_paths, all_metadata_paths):
                process_single_fn = partial(
                    self._process_single_and_save_status,
                    queue=pbar_queue,
                    source_path=s,
                    destination_path=d,
                    metadata_path=m,
                    serialized_kwargs=pickle.dumps(process_single_kwargs),
                )
                result = pool.apply_async(process_single_fn)
                results.append(result)

            for result in results:
                result.get()

            pool.close()
            pool.join()

            pbar_queue.put(None)
            thread.join()
            manager.shutdown()

    def _valid_path(self, path: str) -> bool:
        if self.include_paths is not None and path not in self.include_paths:
            return False
        if self.exclude_paths is not None and path in self.exclude_paths:
            return False
        if self.files_regex_pattern is not None and not self.files_regex_pattern.search(path):
            return False
        return True

    def _get_all_paths(self) -> Tuple[List[str], List[str], List[str]]:
        """Get all paths to process using prefixes provided"""
        all_source_paths, all_destination_paths, all_metadata_paths = [], [], []

        for src_prefix, dst_prefix, meta_prefix in zip(self.src_prefixes, self.dst_prefixes, self.meta_prefixes):
            mkdir_p(dst_prefix)
            mkdir_p(meta_prefix)

            prefix, rel_paths = make_relative(list(glob_path(src_prefix)))
            random.shuffle(rel_paths)

            existing_metadata_names = set(
                sub_prefix(path, meta_prefix).strip(METADATA_SUFFIX) for path in glob_path(meta_prefix)
            )

            for path in rel_paths:
                if not self.ignore_existing and path in existing_metadata_names:
                    continue

                if not self._valid_path(path):
                    continue

                all_source_paths.append(add_suffix(prefix, path))
                all_destination_paths.append(add_suffix(dst_prefix, path))
                all_metadata_paths.append(add_suffix(meta_prefix, path) + METADATA_SUFFIX)

        return all_source_paths, all_destination_paths, all_metadata_paths

    def __call__(self, **process_single_kwargs: Any):
        """Run the processor."""
        random.seed(self.seed)

        # in case the user wants to override the default kwargs for retries
        process_single_kwargs.setdefault("retries_on_error", self.retries_on_error)

        all_source_paths, all_destination_paths, all_metadata_paths = self._get_all_paths()

        print(f"Found {len(all_source_paths):,} files to process")

        fn = self._debug_run_all if self.debug else self._multiprocessing_run_all

        fn(
            all_source_paths=all_source_paths,
            all_destination_paths=all_destination_paths,
            all_metadata_paths=all_metadata_paths,
            **process_single_kwargs,
        )



# flake8: noqa
# type: ignore

import argparse
import json
import os
from contextlib import ExitStack
from typing import Dict, List, Optional

import msgspec
import yaml

from .data_types import DocResult, InputSpec, OutputSpec


class Visualizer:
    BASE_S3_PREFIX = "s3://ai2-llm/pretraining-data/sources"

    def __init__(
        self,
        dataset: str,
        experiment: Optional[str] = None,
        tagger: Optional[str] = None,
        type: Optional[str] = None,
    ):
        self.dataset = dataset
        self.experiment = experiment
        self.tagger = tagger
        self.type = type

    def list_tags(self, path: str):
        prefix, doc_path = path.split("/documents/")

        attrs_decoder = msgspec.json.Decoder(OutputSpec)
        doc_decoder = msgspec.json.Decoder(InputSpec)

        with ExitStack() as stack:
            doc_file = stack.enter_context(stream_file_for_read(path, "rb"))
            doc_stream = stack.enter_context(decompress_stream(doc_file, "rt"))
            exp_path = f"{prefix}/attributes/{self.experiment}/{doc_path}"
            exp_file = stack.enter_context(stream_file_for_read(exp_path, "rb"))
            exp_stream = stack.enter_context(decompress_stream(exp_file, "rt"))

            tags: Dict[str, List[str]] = {}
            for doc_line, exp_line in zip(doc_stream, exp_stream):
                # parse out data from the line
                input_doc = doc_decoder.decode(doc_line)
                input_exp = attrs_decoder.decode(exp_line)
                doc_result = DocResult.from_spec(input_doc, input_exp)

                for span in doc_result.spans:
                    tags.setdefault(str(span.tagger), []).append(span.type)

                break

            print(colored(f"from {self.short_path(path)}:", color="yellow"))
            for tagger, types in sorted(tags.items()):
                print(colored(f"{tagger}:", color="magenta"))
                for type in sorted(set(types)):
                    print(colored(f"  {type}", color="cyan"))
                print()

    def short_path(self, path: str, slack: int = 20) -> str:
        return f"...{path[-s:]}" if (s := round(os.get_terminal_size().columns - slack)) < len(path) else path

    def visualize_single(self, path: str):
        prefix, doc_path = path.split("/documents/")

        attrs_decoder = msgspec.json.Decoder(OutputSpec)
        doc_decoder = msgspec.json.Decoder(InputSpec)

        with ExitStack() as stack:
            doc_file = stack.enter_context(stream_file_for_read(path, "rb"))
            doc_stream = stack.enter_context(decompress_stream(doc_file, "rt"))
            exp_path = f"{prefix}/attributes/{self.experiment}/{doc_path}"
            exp_file = stack.enter_context(stream_file_for_read(exp_path, "rb"))
            exp_stream = stack.enter_context(decompress_stream(exp_file, "rt"))

            i = 0
            file_header = colored(f"file:    {self.short_path(path)}\n", color="magenta")

            for doc_line, exp_line in zip(doc_stream, exp_stream):
                # parse out data from the line
                input_doc = doc_decoder.decode(doc_line)
                input_exp = attrs_decoder.decode(exp_line)
                doc_result = DocResult.from_spec(input_doc, input_exp)

                example_header = colored(f"example: {i:,}\n", color="yellow")
                dt = doc_result.doc.text

                spans = sorted(
                    (s for s in doc_result.spans if s.tagger == self.tagger and s.type == self.type),
                    key=lambda s: s.start,
                )
                if not spans:
                    continue

                prev_start = 0
                text_fragments = []
                for span in spans:
                    text_fragments.append(colored(dt[prev_start : span.start].replace("\n", "\\n"), color="black"))
                    text_fragments.append(colored(dt[span.start : span.end].replace("\n", "\\n"), color="green"))
                    text_fragments.append(colored(f"{{{span.type}: {span.score}}}", color="red"))
                    prev_start = span.end
                text_fragments.append(colored(dt[prev_start:].replace("\n", "\\n"), color="black"))

                tagger_header = colored(f"tagger:  {self.tagger}\n", color="cyan")
                print("".join(text_fragments) + "\n" + file_header + example_header + tagger_header + "\n")
                while True:
                    out = input("next? [l/f] ").lower().strip()
                    if out == "l":
                        i += 1
                        break
                    elif out == "f":
                        return
                    else:
                        print(f"invalid input: {out}; choose between next (l)ine or next (f)ile.")

    def __call__(self):
        try:
            source_prefix = f"{self.BASE_S3_PREFIX}/{self.dataset}/documents"
            for path in recursively_list_files(source_prefix):
                # just list tags if no tagger or type is specified
                if self.tagger is None or self.type is None:
                    self.list_tags(path)
                    return

                # visualize the specified tagger and type
                self.visualize_single(path)
        except KeyboardInterrupt:
            print("\nExiting... bye!")

    @classmethod
    def main(cls):
        ap = argparse.ArgumentParser()
        ap.add_argument(
            "-d",
            "--dataset",
            required=True,
            help="Dataset to visualize",
        )
        ap.add_argument(
            "-e",
            "--experiment-name",
            required=True,
            help="Experiment name to visualize",
        )
        ap.add_argument(
            "-t",
            "--tagger",
            type=str,
            default=None,
            help="Tagger to visualize",
        )
        ap.add_argument(
            "-y",
            "--type",
            type=str,
            default=None,
            help="Type to visualize",
        )
        opts = ap.parse_args()
        cls(dataset=opts.dataset, experiment=opts.experiment_name, tagger=opts.tagger, type=opts.type)()


class RawPreviewer:
    BASE_S3_PREFIX = "s3://ai2-llm/pretraining-data/sources"

    def __init__(self, dataset: str, type: str, file: str, pretty: bool = False, experiment: Optional[str] = None):
        self.dataset = dataset
        self.experiment = experiment
        self.type = type
        self.file = file
        self.pretty = pretty

        assert type == "documents" or experiment is not None, "Must specify experiment for attributes"

    def preview_file(self):
        if self.type == "documents":
            path = f"{self.BASE_S3_PREFIX}/{self.dataset}/documents/{self.file}"
        else:
            path = f"{self.BASE_S3_PREFIX}/{self.dataset}/attributes/{self.experiment}/{self.file}"

        with ExitStack() as stack:
            file = stack.enter_context(stream_file_for_read(path, "rb"))
            stream = stack.enter_context(decompress_stream(file, "rt"))

            list_colors = ["red", "green", "blue", "magenta", "cyan"]

            for line in stream:
                row = json.loads(line)
                if self.pretty:
                    out = yaml.dump(row, width=float("inf"))
                    for ln in out.split("\n"):
                        if not ln.startswith("  "):
                            key, *rest = ln.split(":")
                            rest = (":" + ":".join(rest) if rest else "").strip()
                            print(colored(key, color=list_colors[0]) + rest)
                            list_colors = list_colors[1:] + list_colors[:1]
                        else:
                            print(ln)
                else:
                    print(row)
                input(colored("\n[press enter for next line]", color="yellow"))

    def list_files(self):
        prefix = f"{self.BASE_S3_PREFIX}/{self.dataset}/documents"
        for path in recursively_list_files(prefix):
            print(path[len(prefix) + 1 :])

    def __call__(self):
        try:
            if self.file is not None:
                self.preview_file()
            else:
                self.list_files()
        except KeyboardInterrupt:
            print("\nExiting... bye!")

    @classmethod
    def main(cls):
        ap = argparse.ArgumentParser()
        ap.add_argument("-d", "--dataset", required=True, help="Dataset to preview, e.g. `wikipedia/v0`")
        ap.add_argument(
            "-t",
            "--type",
            choices=["documents", "attributes"],
            required=True,
            help="Type of data to preview; it can be either `documents` or `attributes`.",
        )
        ap.add_argument(
            "-e",
            "--experiment",
            default=None,
            help="Experiment to preview; this is only used for previewing `attributes` assigned by tagger.",
        )
        ap.add_argument(
            "-f",
            "--file",
            default=None,
            type=str,
            help="File to preview; if not sure which file to preview, skip this argument to list all files.",
        )
        ap.add_argument(
            "-p", "--pretty", action="store_true", help="Whether to use pretty print for previewing JSON lines."
        )
        opts = ap.parse_args()

        cls(dataset=opts.dataset, type=opts.type, file=opts.file, experiment=opts.experiment, pretty=opts.pretty)()



import re
import string
from typing import List

try:
    import blingfire

    BLINGFIRE_AVAILABLE = True
except Exception:
    BLINGFIRE_AVAILABLE = False

import nltk
from nltk.tokenize.punkt import PunktSentenceTokenizer

try:
    nltk.data.find("tokenizers/punkt")
except LookupError:
    nltk.download("punkt")


from .data_types import TextSlice

sent_tokenizer = PunktSentenceTokenizer()


def make_variable_name(name: str, remove_multiple_underscores: bool = False) -> str:
    # use underscores for any non-valid characters in variable name
    name = re.sub(r"[^a-zA-Z0-9_]", "_", name)

    if remove_multiple_underscores:
        # replace multiple underscores with a single underscore
        name = re.sub(r"__+", "_", name)

    if name[0] in string.digits:
        raise ValueError(f"Invalid variable name {name}")

    return name


def split_paragraphs(text: str, remove_empty: bool = True) -> List[TextSlice]:
    """
    Split a string into paragraphs. A paragraph is defined as a sequence of zero or more characters, followed
    by a newline character, or a sequence of one or more characters, followed by the end of the string.
    """
    text_slices = [
        TextSlice(doc=text, start=match.start(), end=match.end())
        for match in re.finditer(r"([^\n]*\n|[^\n]+$)", text)
    ]
    if remove_empty is True:
        text_slices = [text_slice for text_slice in text_slices if text_slice.text.strip()]
    return text_slices


def split_sentences(text: str, remove_empty: bool = True) -> List[TextSlice]:
    """
    Split a string into sentences.
    """
    if text and BLINGFIRE_AVAILABLE:
        _, offsets = blingfire.text_to_sentences_and_offsets(text)
    elif text:
        offsets = [(start, end) for start, end in sent_tokenizer.span_tokenize(text)]
    else:
        offsets = []

    if remove_empty is True:
        return [TextSlice(doc=text, start=start, end=end) for (start, end) in offsets]
    else:
        raise NotImplementedError("remove_empty=False is not implemented yet")



from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from omegaconf import OmegaConf as om

from dolma import deduper
from dolma.cli import BaseCli, field, print_config
from dolma.cli.shared import WorkDirConfig


@dataclass
class ParagraphDedupeConfig:
    attribute_name: str = field(help="Name of the output field in the tagger")


@dataclass
class DocumentDedupeConfig:
    attribute_name: str = field(help="Name of the output field in the tagger")
    key: str = field(help="Name of the input field to use for deduplication, e.g. `$.metadata.url`")


@dataclass
class BloomFilterConfig:
    file: str = field(help="Path where to read/write the bloom filter file to/from. Required.")
    size_in_bytes: int = field(
        default=-1,
        help=(
            "Size of the bloom filter in bytes. Either this value is provided, or both estimated_doc_count "
            "and desired_false_positive_rate."
        ),
    )
    read_only: bool = field(help="If true, the bloom filter will be read from the file and not updated. Required.")
    estimated_doc_count: int = field(
        default=-1,
        help=(
            "Estimated number of documents to be added to the bloom filter. Either this value is provided, "
            "or both size_in_bytes and desired_false_positive_rate."
        ),
    )
    desired_false_positive_rate: float = field(
        default=-1.0,
        help=(
            "Desired false positive rate. Either this value is provided, or both size_in_bytes and "
            "estimated_doc_count."
        ),
    )


@dataclass
class DedupeConfig:
    name: str = field(help="Name of the deduper. Required.")
    documents: Optional[DocumentDedupeConfig] = field(
        default=None, help="Configuration for document deduplication"
    )
    paragraphs: Optional[ParagraphDedupeConfig] = field(
        default=None, help="Configuration for paragraph deduplication"
    )
    skip_empty: Optional[bool] = field(default=False, help="If true, empty documents/paragraphs will be skipped")


@dataclass
class DeduperConfig:
    documents: List[str] = field(default=[], help="Paths to the documents to be deduplicated. Required.")
    work_dir: WorkDirConfig = field(default=WorkDirConfig(), help="Configuration for temporary work directories.")
    dedupe: DedupeConfig = field(help="Deduplication configuration. Required.")
    bloom_filter: BloomFilterConfig = field(help="Bloom filter configuration. Required.")
    processes: int = field(
        default=1, help="Number of processes to use for deduplication. If 1, no multiprocessing will be used."
    )


class DeduperCli(BaseCli):
    CONFIG = DeduperConfig

    @classmethod
    def run(cls, parsed_config: DeduperConfig):
        dict_config: Dict[str, Any] = {}

        dict_config["dedupe"] = {"name": parsed_config.dedupe.name, "skip_empty": parsed_config.dedupe.skip_empty}
        if parsed_config.dedupe.documents is not None:
            dict_config["dedupe"]["documents"] = om.to_container(parsed_config.dedupe.documents)
        elif parsed_config.dedupe.paragraphs is not None:
            dict_config["dedupe"]["paragraphs"] = om.to_container(parsed_config.dedupe.paragraphs)
        else:
            raise ValueError("Either dedupe.documents or dedupe.paragraphs must be specified")

        dict_config["bloom_filter"] = {
            "file": parsed_config.bloom_filter.file,
            "read_only": parsed_config.bloom_filter.read_only,
            "size_in_bytes": getattr(parsed_config.bloom_filter, "size_in_bytes", 0),
            "estimated_doc_count": getattr(parsed_config.bloom_filter, "estimated_doc_count", 0),
            "desired_false_positive_rate": getattr(parsed_config.bloom_filter, "desired_false_positive_rate", 0),
        }

        if dict_config["bloom_filter"]["size_in_bytes"] <= 0 and (
            dict_config["bloom_filter"]["estimated_doc_count"] <= 0
            or dict_config["bloom_filter"]["desired_false_positive_rate"] <= 0
        ):
            raise ValueError(
                "Either bloom_filter.size_in_bytes or bloom_filter.estimated_doc_count and "
                "bloom_filter.desired_false_positive_rate must be specified"
            )

        dict_config["work_dir"] = {"input": parsed_config.work_dir.input, "output": parsed_config.work_dir.output}
        dict_config["processes"] = parsed_config.processes
        dict_config["documents"] = list(om.to_container(parsed_config.documents))  # pyright: ignore

        if len(dict_config["documents"]) == 0:
            raise ValueError("At least one document must be specified")

        print_config(dict_config)
        return deduper(dict_config)



import logging
import os
from dataclasses import dataclass
from typing import List, Set

from ..core.data_types import DocResult, Document, Span
from ..core.registry import TaggerRegistry
from ..core.taggers import BaseTagger

MIN_WORDS_PER_LINE = 3
naughty_lines = (
    open(os.path.join(os.path.dirname(os.path.dirname(__file__)), "data", "naughty_words_en.txt"))
    .read()
    .splitlines()
)
NAUGHTY_WORDS: Set[str] = set(w for w in naughty_lines if " " not in w)
NAUGHTY_PHRASES: Set[str] = set(w for w in naughty_lines if " " in w)


@dataclass
class C4Attributes:
    lines_with_no_ending_punctuation: List[Span]
    lines_with_too_few_words: List[Span]
    has_naughty_word: bool = False
    has_javascript: bool = False
    has_lorem_ipsum: bool = False
    has_curly_brace: bool = False
    line_count: int = 0
    character_count: int = 0

    def as_spans(self) -> List[Span]:
        spans = []
        spans.extend(self.lines_with_no_ending_punctuation)
        spans.extend(self.lines_with_too_few_words)
        if self.has_naughty_word:
            spans.append(Span(0, self.character_count, type="has_naughty_word"))
        if self.has_javascript:
            spans.append(Span(0, self.character_count, type="has_javascript"))
        if self.has_lorem_ipsum:
            spans.append(Span(0, self.character_count, type="has_lorem_ipsum"))
        if self.has_curly_brace:
            spans.append(Span(0, self.character_count, type="has_curly_brace"))
        spans.append(Span(0, self.character_count, type="line_count", score=self.line_count))
        return spans


def get_attributes(text: str) -> C4Attributes:
    attrs = C4Attributes([], [])
    attrs.character_count = len(text)
    try:
        lines = text.split("\n")
        attrs.line_count = len(lines)
        offset = 0
        for line_no in range(0, len(lines)):
            original_line = lines[line_no]
            end_offset = offset + len(original_line)
            if line_no < len(lines) - 1:
                end_offset += 1
            line = original_line.lower().strip()
            if not line.endswith((".", "?", "!", '"')):
                attrs.lines_with_no_ending_punctuation.append(
                    Span(offset, end_offset, type="lines_with_no_ending_punctuation")
                )
            words = line.split()
            if len(words) < MIN_WORDS_PER_LINE:
                attrs.lines_with_too_few_words.append(Span(offset, end_offset, type="lines_with_too_few_words"))
            if any(word in NAUGHTY_WORDS for word in words) or any(phrase in line for phrase in NAUGHTY_PHRASES):
                attrs.has_naughty_word = True
            if any(word == "javascript" for word in words):
                attrs.has_javascript = True
            if "lorem ipsum" in line:
                attrs.has_lorem_ipsum = True
            if "{" in line:
                attrs.has_curly_brace = True
            offset = end_offset
    except Exception:
        logging.exception(f"Error parsing text: {text[:200]}")

    return attrs


@TaggerRegistry.add("c4_v1")
class C4Tagger(BaseTagger):
    def predict(self, doc: Document) -> DocResult:
        attrs = get_attributes(doc.text)
        result = DocResult(doc=doc, spans=attrs.as_spans())
        return result



"""

Filters.

@kylel, @soldni

"""

try:
    import re2 as re
except ImportError:
    import re
else:
    re.set_fallback_notification(re.FALLBACK_WARNING)


from typing import List
from warnings import warn

from presidio_analyzer import AnalyzerEngine

from ..core.data_types import DocResult, Document, Span, TextSlice
from ..core.registry import TaggerRegistry
from ..core.taggers import BaseTagger
from ..core.utils import split_paragraphs

__all__ = ["PiiPresidioV1", "PiiRegexV1", "PiiRegexV2", "FastPiiRegex", "PiiRegexWithCountV2"]


class BasePiiFilter(BaseTagger):
    EMAIL = "EMAIL_ADDRESS"
    PHONE = "PHONE_NUMBER"
    IP = "IP_ADDRESS"

    PRESIDIO = "presidio"
    REGEX = "regex"

    ENGLISH = "en"
    WINDOW = 100

    def __init__(
        self,
        method: str,
        postprocess: bool,
        window: int,
    ) -> None:
        assert method in [
            self.PRESIDIO,
            self.REGEX,
        ], f"Please provide a valid method for filtering ({self.PRESIDIO} or {self.REGEX})"

        # configs
        self.method = method
        self.postprocess = postprocess
        self.window = window

        # Regular expressions for different types of PII
        self.pii_type_to_regex = {
            self.EMAIL: re.compile("[.\\s@,?!;:)(]*([^\\s@]+@[^\\s@,?!;:)(]+?)[.\\s@,?!;:)(]?[\\s\n\r]"),
            self.PHONE: re.compile("\\s+\\(?(\\d{3})\\)?[-\\. ]*(\\d{3})[-. ]?(\\d{4})"),
            self.IP: re.compile(
                "(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
            ),
        }
        self.url_regex = re.compile(
            "(?i)\b((?:https?://|www\\d{0,3}[.]|[a-z0-9.\\-]+[.][a-z]{2,4}/)(?:[^\\s()<>]+|\\(([^\\s()<>]+|"
            "(\\([^\\s()<>]+\\)))*\\))+(?:\\(([^\\s()<>]+|(\\([^\\s()<>]+\\)))*\\)|[^\\s`!()\\[\\]"
            "{};:'\".,<>?«»“”‘’]))"
        )

        # presidio
        if self.method == self.PRESIDIO:
            self.analyzer = AnalyzerEngine()

    def predict(self, doc: Document) -> DocResult:
        """Main runner."""
        # extract
        if self.method == self.PRESIDIO:
            pii_spans = self._extract_pii_presidio(text=doc.text)
        elif self.method == self.REGEX:
            pii_spans = self._extract_pii_regex(text=doc.text)
        else:
            raise NotImplementedError
        # post process
        if self.postprocess:
            new_pii_spans = self._postprocess(text=doc.text, pii_spans=pii_spans, window=self.window)
        else:
            new_pii_spans = pii_spans

        # document-level score
        score = self._score(text=doc.text, pii_spans=new_pii_spans)
        new_pii_spans.append(Span(start=0, end=len(doc.text), type="doc", score=score))
        return DocResult(doc=doc, spans=new_pii_spans)

    def _score(self, text: str, pii_spans: List[Span]) -> float:
        return len(pii_spans) * 1.0 / len(text.split())

    def _extract_pii_regex(self, text: str) -> List[Span]:
        pii_spans: List[Span] = []
        for pii_type, regex in self.pii_type_to_regex.items():
            for match in regex.finditer(text):
                start, end = match.span()
                pii_spans.append(Span(start=start, end=end, type=pii_type))
        return pii_spans

    def _extract_pii_presidio(self, text: str) -> List[Span]:
        analyzer_results = self.analyzer.analyze(
            text=text,
            entities=[self.EMAIL, self.PHONE, self.IP],
            language=self.ENGLISH,
        )
        pii_spans: List[Span] = []
        for res in analyzer_results:
            pii_spans.append(Span(start=res.start, end=res.end, type=res.entity_type))
        return pii_spans

    def _postprocess(self, text: str, pii_spans: List[Span], window: int) -> List[Span]:
        """Applies some rules to remove over-prediction of PII types."""
        new_pii_spans = []
        for pii_span in pii_spans:
            if pii_span.type == self.EMAIL:
                if self._is_email(text, pii_span):
                    new_pii_spans.append(pii_span)
                else:
                    pass

            elif pii_span.type == self.PHONE or pii_span.type == self.IP:
                context = pii_span.mention(text=text, window=window)
                # for both phone numbers & IP addresses, context shouldnt
                # contain these strings
                if "isbn" in context or "doi" in context or "#" in context:
                    pass
                elif pii_span.type == self.IP:
                    new_pii_spans.append(pii_span)
                elif pii_span.type == self.PHONE:
                    # for phone numbers, additionally shouldnt be URL
                    if self._contains_url(text=text):
                        pass
                    else:
                        new_pii_spans.append(pii_span)

            else:
                raise NotImplementedError(f"Unsupported PII type for Postprocess: {pii_span.type}")
        return new_pii_spans

    def _contains_url(self, text: str) -> bool:
        return len(self.url_regex.findall(text)) > 0

    def _is_email(self, text: str, pii_span: Span) -> bool:
        """
        Rules:
        (1) The email address besides the domain, cannot be only "("
        (2) There must be a "." in the domain
        """
        mention = pii_span.mention(text=text)
        addressee = mention.split("@")[0]
        domain = mention.split("@")[1]
        if addressee.strip() == "(" or "." not in domain:
            return False
        return True


@TaggerRegistry.add("pii_presidio_v1")
class PiiPresidioV1(BasePiiFilter):
    def __init__(self):
        super().__init__(method=self.PRESIDIO, postprocess=True, window=self.WINDOW)


@TaggerRegistry.add("pii_regex_v1")
class PiiRegexV1(BasePiiFilter):
    def __init__(self):
        super().__init__(method=self.REGEX, postprocess=True, window=self.WINDOW)


@TaggerRegistry.add("pii_regex_v2")
class PiiRegexV2(PiiRegexV1):
    def _score(self, text: str, pii_spans: List[Span]) -> float:
        try:
            score = len(pii_spans) * 1.0 / len(text.split())
        except ZeroDivisionError:
            score = -1.0
        return score


@TaggerRegistry.add("pii_regex_with_counts_fast_v2")
class FastPiiRegex(BaseTagger):
    EMAIL_KEY = "EMAIL_ADDRESS"
    PHONE_KEY = "PHONE_NUMBER"
    IP_KEY = "IP_ADDRESS"

    EMAIL_REGEX = "[.\\s@,?!;:)(]*([^\\s@]+@[^\\s@,?!;:)(]+?)[.\\s@,?!;:)(]?[\\s\n\r]"
    PHONE_REGEX = "\\s+\\(?(\\d{3})\\)?[-\\. ]*(\\d{3})[-. ]?(\\d{4})"
    IP_REGEX = "(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
    URL_REGEX = "(?i)\b((?:https?://|www\\d{0,3}[.]|[a-z0-9.\\-]+[.][a-z]{2,4}/)(?:[^\\s()<>]+|\\(([^\\s()<>]+|(\\([^\\s()<>]+\\)))*\\))+(?:\\(([^\\s()<>]+|(\\([^\\s()<>]+\\)))*\\)|[^\\s`!()\\[\\]{};:'\".,<>?«»“”‘’]))"  # noqa: E501

    def __init__(
        self,
        email_regex: str = EMAIL_REGEX,
        phone_regex: str = PHONE_REGEX,
        ip_regex: str = IP_REGEX,
        url_regex: str = URL_REGEX,
    ) -> None:
        self.email_regex = re.compile(email_regex)
        self.phone_regex = re.compile(phone_regex)
        self.ip_regex = re.compile(ip_regex)
        self.url_regex = re.compile(url_regex)

        self.pre_ip_regex = re.compile(r"\.[^\s]")
        self.pre_phone_regex = re.compile(r"\d")

    def _false_positive_identifiers(self, text: str) -> bool:
        return "isbn" in text or "doi" in text or "#" in text

    def _predict_email(self, slice: TextSlice) -> List[Span]:
        if "@" not in slice.text:
            return []

        spans = []
        for match in self.email_regex.finditer(slice.text):
            addressee, domain = match.group(1).split("@", 1)
            if addressee.strip() == "(" or "." not in domain:
                continue

            start, end = match.span()
            spans.append(Span(start=start + slice.start, end=end + slice.start, type=self.EMAIL_KEY))

        return spans

    def _predict_phone(self, slice: TextSlice) -> List[Span]:
        if not self.pre_phone_regex.search(slice.text):
            return []

        spans = []
        for match in self.phone_regex.finditer(slice.text):
            start, end = match.span()
            spans.append(Span(start=start + slice.start, end=end + slice.start, type=self.PHONE_KEY))

        return spans

    def _predict_ip(self, slice: TextSlice) -> List[Span]:
        if not self.pre_ip_regex.search(slice.text):
            return []

        spans = []
        for match in self.ip_regex.finditer(slice.text):
            if self._contains_url(match.group(0)):
                continue
            start, end = match.span()
            spans.append(Span(start=start + slice.start, end=end + slice.start, type=self.IP_KEY))

        return spans

    def _contains_url(self, text: str) -> bool:
        return self.url_regex.search(text) is not None

    def predict(self, doc: Document) -> DocResult:
        paragraphs = split_paragraphs(doc.text)
        spans: List[Span] = []

        if doc.text.count("?") > 10_000:
            warn("Skipping regex PII detection for doc with >10k question marks")
            paragraphs = []

        for paragraph in paragraphs:
            spans.extend(self._predict_email(paragraph))
            spans.extend(self._predict_phone(paragraph))
            spans.extend(self._predict_ip(paragraph))

        # doc level score is the count of spans matching any of the PII types
        score = sum(1.0 for s in spans if s.type != "doc")
        spans.append(Span(start=0, end=len(doc.text), type="doc_count", score=score))

        try:
            # fraction of words that are PII
            score = sum(len(s) for s in spans) / len(doc.text)
        except ZeroDivisionError:
            # empty doc
            score = -1.0

        spans.append(Span(start=0, end=len(doc.text), type="doc_frac", score=score))
        return DocResult(doc=doc, spans=spans)


@TaggerRegistry.add("pii_regex_with_counts_v2")
class PiiRegexWithCountV2(BasePiiFilter):
    def __init__(self):
        super().__init__(method=self.REGEX, postprocess=True, window=self.WINDOW)

    def predict(self, doc: Document) -> DocResult:
        doc_result = super().predict(doc=doc)
        count = sum(1 for s in doc_result.spans if s.type != "doc")
        doc_result.spans.append(Span(start=0, end=len(doc.text), type="doc_count", score=count))
        return doc_result



from argparse import ArgumentParser
from dataclasses import dataclass
from unittest import TestCase

from omegaconf import MissingMandatoryValue

from dolma.cli import field, make_parser, namespace_to_nested_omegaconf


@dataclass
class _1:
    a: int = field(help="a")
    b: str = field(help="b")


@dataclass
class _2:
    a: _1 = field(help="a")
    c: float = field(help="c", default=1.0)


class TestOmegaconf(TestCase):
    def test_make_parser(self):
        ap = ArgumentParser()
        parser = make_parser(ap, _1)

        args = parser.parse_args(["--a", "1", "--b", "2"])
        opts = namespace_to_nested_omegaconf(args=args, structured=_1)

        self.assertEqual(opts.a, 1)
        self.assertEqual(opts.b, "2")

    def test_nested_parser(self):
        ap = ArgumentParser()
        parser = make_parser(ap, _2)

        args = parser.parse_args(["--a.a", "1", "--a.b", "2", "--c", "3"])
        opts = namespace_to_nested_omegaconf(args=args, structured=_2)

        self.assertEqual(opts.a.a, 1)
        self.assertEqual(opts.a.b, "2")
        self.assertEqual(opts.c, 3.0)

    def test_fail_omegaconf(self):
        ap = ArgumentParser()
        parser = make_parser(ap, _2)

        args = parser.parse_args(["--a.a", "1", "--c", "3"])
        conf = namespace_to_nested_omegaconf(args, _2)

        with self.assertRaises(MissingMandatoryValue):
            conf.a.b



from dataclasses import dataclass
from typing import List, Optional

from omegaconf import MISSING

from dolma.cli import BaseCli, field, print_config
from dolma.cli.shared import WorkDirConfig
from dolma.core.runtime import create_and_run_tagger


@dataclass
class TaggerConfig:
    documents: List[str] = field(
        default=[],
        help="One or more document paths to process; Can be either local or S3 paths. Globs are supported.",
    )
    destination: Optional[List[str]] = field(
        default=None,
        help=(
            "Destination paths to save the outputs; should match the number of document paths. "
            "If not provided, destination will be derived from the document path."
        ),
    )
    taggers: List[str] = field(
        default=[],
        help="List of taggers to run.",
    )
    experiment: str = field(
        default=MISSING,
        help="Name of the experiment.",
    )
    processes: int = field(
        default=1,
        help="Number of parallel processes to use.",
    )
    debug: bool = field(
        default=False,
        help="Whether to run in debug mode.",
    )
    work_dir: Optional[WorkDirConfig] = field(
        default=WorkDirConfig(), help="Configuration for temporary work directories."
    )


class TaggerCli(BaseCli):
    CONFIG = TaggerConfig

    @classmethod
    def run(cls, parsed_config: TaggerConfig):
        metadata = parsed_config.work_dir.output if parsed_config.work_dir else None
        documents = [str(p) for p in parsed_config.documents]
        taggers = [str(p) for p in parsed_config.taggers]

        print_config(parsed_config)
        create_and_run_tagger(
            documents=documents,
            destination=parsed_config.destination,
            metadata=metadata,
            taggers=taggers,
            num_processes=parsed_config.processes,
            experiment=parsed_config.experiment,
            debug=parsed_config.debug,
        )



"""

Filters.

@kylel, @soldni

"""
from abc import abstractmethod
from typing import Dict, List, Union

from .data_types import DocResult, Document, InputSpec


class BaseTagger:
    @classmethod
    def train(cls, *args, **kwargs):
        raise RuntimeError("This tagger does not support training")

    @classmethod
    def test(cls, *args, **kwargs):
        raise RuntimeError("This tagger does not support testing")

    @abstractmethod
    def predict(self, doc: Document) -> DocResult:
        raise NotImplementedError

    def tag(self, row: InputSpec) -> Dict[str, List[List[Union[int, float]]]]:
        """Internal function that is used by the tagger to get data"""
        doc = Document(source=row.source, version=row.version, id=row.id, text=row.text)
        doc_result = self.predict(doc)

        tagger_output: Dict[str, list] = {}
        for span in doc_result.spans:
            tagger_output.setdefault(span.type, []).append([span.start, span.end, round(float(span.score), 5)])
        return tagger_output



"""

Filters.

@kylel, @soldni

"""

import regex
import uniseg.wordbreak
from tokenizers import Regex, pre_tokenizers

from ..core.data_types import DocResult, Document, Span
from ..core.registry import TaggerRegistry
from ..core.taggers import BaseTagger
from ..core.utils import split_paragraphs


@TaggerRegistry.add("char_length_v1")
class CharLengthV1(BaseTagger):
    def predict(self, doc: Document) -> DocResult:
        score = len(doc.text)
        return DocResult(doc=doc, spans=[Span(start=0, end=len(doc.text), type="length", score=score)])


@TaggerRegistry.add("char_length_with_paragraphs_v1")
class CharLengthWithParagraphsV1(BaseTagger):
    def predict(self, doc: Document) -> DocResult:
        spans = [
            Span(start=p.start, end=p.end, type="paragraph", score=len(p.text)) for p in split_paragraphs(doc.text)
        ]
        spans.append(Span(start=0, end=len(doc.text), type="document", score=len(doc.text)))
        return DocResult(doc=doc, spans=spans)


@TaggerRegistry.add("whitespace_tokenizer_v1")
class WhitespaceLengthV1(BaseTagger):
    WHITESPACE_REGEX = regex.compile(r"\w+|[^\w\s]+")

    def predict(self, doc: Document) -> DocResult:
        score = len(self.WHITESPACE_REGEX.split(doc.text))
        return DocResult(doc=doc, spans=[Span(start=0, end=len(doc.text), type="length", score=score)])


@TaggerRegistry.add("whitespace_tokenizer_with_paragraphs_v1")
class WhitespaceLengthParagraphsV1(WhitespaceLengthV1):
    def predict(self, doc: Document) -> DocResult:
        spans = [
            Span(start=p.start, end=p.end, type="paragraph", score=len(self.WHITESPACE_REGEX.split(p.text)))
            for p in split_paragraphs(doc.text)
        ]
        spans.append(Span(start=0, end=len(doc.text), type="document", score=sum(s.score for s in spans)))
        return DocResult(doc=doc, spans=spans)


@TaggerRegistry.add("uniseg_length_paragraphs_v1")
class UnisegParagraphsV1(BaseTagger):
    def predict(self, doc: Document) -> DocResult:
        spans = []
        for para in split_paragraphs(doc.text):
            # we ignore whitespace-only tokens when counting words
            para_length = sum(1 for w in uniseg.wordbreak.words(para.text.strip()) if w.strip())
            spans.append(Span(start=para.start, end=para.end, type="paragraph", score=para_length))

            # we have to record negative length because mixer can only work on greater than operations,
            # and we might want to drop paragraphs that are shorter than a certain length n, so we need
            # to filter on >= n
            spans.append(Span(start=para.start, end=para.end, type="negative_paragraph", score=-para_length))
        return DocResult(doc=doc, spans=spans)


@TaggerRegistry.add("uniseg_length_paragraphs_with_doc_length_v1")
class UnisegParagraphsWithDocLengthV1(UnisegParagraphsV1):
    def predict(self, doc: Document) -> DocResult:
        doc_results = super().predict(doc)
        pos_len = sum(s.score for s in doc_results.spans if s.type == "paragraph")
        neg_len = sum(s.score for s in doc_results.spans if s.type == "negative_paragraph")
        doc_results.spans.append(Span(start=0, end=len(doc.text), type="document", score=pos_len))
        doc_results.spans.append(Span(start=0, end=len(doc.text), type="negative_document", score=neg_len))
        return doc_results


@TaggerRegistry.add("olmo_pretokenizer_v1")
class OlmoPreTokenizerV1(BaseTagger):
    def __init__(self) -> None:
        self.pre_tokenizer = pre_tokenizers.Sequence(
            [
                # Split on all punctuation.
                pre_tokenizers.Split(
                    pattern=Regex(" ?[[:punct:]]"),
                    behavior="isolated",
                    invert=False,
                ),
                # Split up digits.
                pre_tokenizers.Split(
                    pattern=Regex(" ?\\d"),
                    behavior="isolated",
                    invert=False,
                ),
                pre_tokenizers.ByteLevel(add_prefix_space=False, use_regex=True),
            ]
        )

    def predict(self, doc: Document) -> DocResult:
        score = len(self.pre_tokenizer.pre_tokenize_str(doc.text))
        return DocResult(doc=doc, spans=[Span(start=0, end=len(doc.text), type="length", score=score)])


@TaggerRegistry.add("olmo_pretokenizer_with_paragraphs_v1")
class OlmoPreTokenizerParagraphsV1(OlmoPreTokenizerV1):
    def predict(self, doc: Document) -> DocResult:
        spans = [
            Span(
                start=p.start, end=p.end, type="paragraph", score=len(self.pre_tokenizer.pre_tokenize_str(p.text))
            )
            for p in split_paragraphs(doc.text)
        ]
        spans.append(Span(start=0, end=len(doc.text), type="document", score=sum(s.score for s in spans)))
        return DocResult(doc=doc, spans=spans)



"""

Code-related taggers.

@akshitab

"""
import logging
import re
from typing import Generator, List

import numpy as np
import regex
from detect_secrets import SecretsCollection
from detect_secrets.core.scan import (
    PotentialSecret,
    _process_line_based_plugins,
    get_plugins,
)
from detect_secrets.settings import default_settings

from ..core.data_types import DocResult, Document, Span
from ..core.registry import TaggerRegistry
from ..core.taggers import BaseTagger

logger = logging.getLogger(__name__)


def scan_code(code: str) -> Generator[PotentialSecret, None, None]:
    if not get_plugins():
        logger.error("No plugins to scan with!")
        return

    has_secret = False
    for lines in [code.splitlines()]:
        for secret in _process_line_based_plugins(
            lines=list(enumerate(lines, start=1)),
            filename="code_str.yml",
        ):
            has_secret = True
            yield secret

        if has_secret:
            break


class SecretsCollectionForStringInput(SecretsCollection):
    def scan_str(self, code_str: str):
        for secret in scan_code(code_str):
            self["code_str.yml"].add(secret)


def get_secrets(code: str):
    secrets = SecretsCollectionForStringInput()
    with default_settings():
        secrets.scan_str(code)

    return secrets


@TaggerRegistry.add("code_secrets_v1")
class CodeSecretsTagger(BaseTagger):
    @classmethod
    def _extract_code_secrets(cls, text: str) -> List[Span]:
        secrets_spans: List[Span] = []

        text_lines = text.splitlines()
        secrets = get_secrets(text)
        for _, secret in secrets:
            line_number = secret.line_number - 1
            span = secret.secret_value
            span_line = text_lines[line_number]
            line_start = text.find(span_line)
            start = line_start + span_line.find(span)
            end = start + len(span)
            assert text[start:end] == span
            secret_type = secret.type.replace(" ", "_")
            secrets_spans.append(Span(start=start, end=end, type=f"SECRET_{secret_type}"))  # , span])

        return secrets_spans

    def predict(self, doc: Document) -> DocResult:
        """Main runner."""
        spans = self._extract_code_secrets(doc.text)

        # document-level score
        score = self._score(text=doc.text, secrets_spans=spans)
        spans.append(Span(start=0, end=len(doc.text), type="doc", score=score))
        return DocResult(doc=doc, spans=spans)

    def _score(self, text: str, secrets_spans: List[Span]) -> float:
        try:
            score = len(secrets_spans) * 1.0 / len(text.split())
        except ZeroDivisionError:
            score = -1.0
        return score


@TaggerRegistry.add("code_copyright_comments_v1")
class CodeCopyrightTagger(BaseTagger):
    """
    Based on RedPajama code filtering.
    """

    def __init__(self):
        self.cpat = re.compile("copyright", re.IGNORECASE)
        self.pat = re.compile("/\\*[^*]*\\*+(?:[^/*][^*]*\\*+)*/")

    def _extract_copyright_spans(self, text: str) -> List[Span]:
        copyright_spans: List[Span] = []

        reg = self.pat.search(text)

        if reg:
            # found one, now see if it contains "copyright", if so strip it
            span = reg.span()
            sub = text[span[0] : span[1]]
            if self.cpat.search(sub):
                copyright_spans.append(Span(start=span[0], end=span[1], type="copyright_notice", score=1.0))
            return copyright_spans

        lines = text.split("\n")
        skip = 0
        # Greedy replace any file that begins with comment block, most
        # are copyright headers
        end = 0
        for k in range(len(lines)):
            if lines[k].startswith("//") or lines[k].startswith("#") or lines[k].startswith("--") or not lines[k]:
                skip = skip + 1
                if not lines[k]:
                    end += 1
                else:
                    end += len(lines[k])
            else:
                break

        if skip:
            copyright_spans.append(Span(start=0, end=end, type="comment_block", score=1.0))
        return copyright_spans

    def predict(self, doc: Document) -> DocResult:
        """Main runner."""
        spans = self._extract_copyright_spans(doc.text)

        # document-level score
        score = self._score(text=doc.text, copyright_spans=spans)
        spans.append(Span(start=0, end=len(doc.text), type="doc", score=score))
        return DocResult(doc=doc, spans=spans)

    def _score(self, text: str, copyright_spans: List[Span]) -> float:
        try:
            if len(copyright_spans) == 0:
                score = 0.0
            else:
                span = copyright_spans[0]
                # percentage of content affected
                score = (span.end - span.start + 1) * 1.0 / len(text)
        except ZeroDivisionError:
            score = -1.0
        return score


@TaggerRegistry.add("code_redpajama_taggers_v1")
class CodeRedPajamaTaggers(BaseTagger):
    """
    Based on RedPajama code filtering.
    """

    WHITESPACE_REGEX = regex.compile(r"\w+|[^\w\s]+")

    def _get_num_tokens(self, text: str):
        return len(self.WHITESPACE_REGEX.split(text))

    def predict(self, doc: Document) -> DocResult:
        """Main runner."""

        spans: List[Span] = []

        doc_length = len(doc.text)

        line_lengths = list(map(len, doc.text.splitlines()))

        max_line_length = max(line_lengths, default=0.0)
        avg_line_length = np.mean(line_lengths) if len(line_lengths) > 0 else 0.0

        alnum_count = sum(map(lambda char: 1 if char.isalnum() else 0, doc.text))
        alnum_prop = (alnum_count / doc_length) if doc_length > 0 else 0.0

        num_tokens = self._get_num_tokens(doc.text)
        num_alpha = len([c for c in doc.text if c.isalpha()])
        alpha_token_prop = (num_alpha / num_tokens) if num_tokens > 0 else 0.0

        # document-level scores
        spans.append(Span(start=0, end=doc_length, type="max_line_length_doc", score=max_line_length))
        spans.append(Span(start=0, end=doc_length, type="avg_line_length_doc", score=avg_line_length))
        spans.append(Span(start=0, end=doc_length, type="alnum_prop_doc", score=alnum_prop))
        spans.append(Span(start=0, end=doc_length, type="alpha_token_prop_doc", score=alpha_token_prop))

        return DocResult(doc=doc, spans=spans)



"""

Unit tests for taggers/*.py

@kylel

"""

from unittest import TestCase

from dolma.core.data_types import Document
from dolma.taggers.gopher import GopherTagger


class TestGopherTagger(TestCase):
    def test_predict_short(self):
        tagger = GopherTagger()
        doc = Document(source="", version="", id="", text="This is a test.")
        doc_result = tagger.predict(doc=doc)
        d = doc_result.to_json()
        self.assertEqual(len(d["spans"]), 13)
        self.assertEqual(
            d["spans"][0],
            {
                "start": 0,
                "end": 15,
                "type": "fraction_of_characters_in_most_common_2grams",
                "score": 0.5,
                "mention": "This is a test.",
            },
        )
        self.assertEqual(
            d["spans"][1],
            {
                "start": 0,
                "end": 15,
                "type": "fraction_of_characters_in_most_common_3grams",
                "score": 0.5833333333333334,
                "mention": "This is a test.",
            },
        )
        self.assertEqual(
            d["spans"][2],
            {
                "start": 0,
                "end": 15,
                "type": "fraction_of_characters_in_most_common_4grams",
                "score": 1.0,
                "mention": "This is a test.",
            },
        )
        self.assertEqual(
            d["spans"][3],
            {"start": 0, "end": 15, "type": "character_count", "score": 15.0, "mention": "This is a test."},
        )
        self.assertEqual(
            d["spans"][4],
            {"start": 0, "end": 15, "type": "word_count", "score": 4.0, "mention": "This is a test."},
        )
        self.assertEqual(
            d["spans"][5],
            {"start": 0, "end": 15, "type": "median_word_length", "score": 3.0, "mention": "This is a test."},
        )
        self.assertEqual(
            d["spans"][6],
            {"start": 0, "end": 15, "type": "symbol_to_word_ratio", "score": 0.0, "mention": "This is a test."},
        )
        self.assertEqual(
            d["spans"][7],
            {
                "start": 0,
                "end": 15,
                "type": "fraction_of_words_with_alpha_character",
                "score": 1.0,
                "mention": "This is a test.",
            },
        )
        self.assertEqual(
            d["spans"][8],
            {"start": 0, "end": 15, "type": "required_word_count", "score": 0.0, "mention": "This is a test."},
        )
        self.assertEqual(
            d["spans"][9],
            {
                "start": 0,
                "end": 15,
                "type": "fraction_of_lines_starting_with_bullet_point",
                "score": 0.0,
                "mention": "This is a test.",
            },
        )
        self.assertEqual(
            d["spans"][10],
            {
                "start": 0,
                "end": 15,
                "type": "fraction_of_lines_ending_with_ellipsis",
                "score": 0.0,
                "mention": "This is a test.",
            },
        )
        self.assertEqual(
            d["spans"][11],
            {
                "start": 0,
                "end": 15,
                "type": "fraction_of_duplicate_lines",
                "score": 0.0,
                "mention": "This is a test.",
            },
        )
        self.assertEqual(
            d["spans"][12],
            {
                "start": 0,
                "end": 15,
                "type": "fraction_of_characters_in_duplicate_lines",
                "score": 0.0,
                "mention": "This is a test.",
            },
        )

    def test_predict_multiline(self):
        tagger = GopherTagger()
        text = "This is a sentence. \n  \n This is another sentence.\n\n  This is a third sentence."
        doc = Document(source="", version="", id="", text=text)
        doc_result = tagger.predict(doc=doc)
        d = doc_result.to_json()
        self.assertEqual(len(d["spans"]), 19)
        self.assertEqual(
            d["spans"][0],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_characters_in_most_common_2grams",
                "score": 0.3050847457627119,
                "mention": text,
            },
        )
        self.assertEqual(
            d["spans"][1],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_characters_in_most_common_3grams",
                "score": 0.23728813559322035,
                "mention": text,
            },
        )
        self.assertEqual(
            d["spans"][2],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_characters_in_most_common_4grams",
                "score": 0.2711864406779661,
                "mention": text,
            },
        )
        self.assertEqual(
            d["spans"][3],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_characters_in_duplicate_5grams",
                "score": 0.0,
                "mention": text,
            },
        )
        self.assertEqual(
            d["spans"][4],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_characters_in_duplicate_6grams",
                "score": 0.0,
                "mention": text,
            },
        )
        self.assertEqual(
            d["spans"][5],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_characters_in_duplicate_7grams",
                "score": 0.0,
                "mention": text,
            },
        )
        self.assertEqual(
            d["spans"][6],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_characters_in_duplicate_8grams",
                "score": 0.0,
                "mention": text,
            },
        )
        self.assertEqual(
            d["spans"][7],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_characters_in_duplicate_9grams",
                "score": 0.0,
                "mention": text,
            },
        )
        self.assertEqual(
            d["spans"][8],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_characters_in_duplicate_10grams",
                "score": 0.0,
                "mention": text,
            },
        )
        self.assertEqual(
            d["spans"][9], {"start": 0, "end": 79, "type": "character_count", "score": 79.0, "mention": text}
        )
        self.assertEqual(
            d["spans"][10], {"start": 0, "end": 79, "type": "word_count", "score": 13.0, "mention": text}
        )
        self.assertEqual(
            d["spans"][11], {"start": 0, "end": 79, "type": "median_word_length", "score": 4.0, "mention": text}
        )
        self.assertEqual(
            d["spans"][12], {"start": 0, "end": 79, "type": "symbol_to_word_ratio", "score": 0.0, "mention": text}
        )
        self.assertEqual(
            d["spans"][13],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_words_with_alpha_character",
                "score": 1.0,
                "mention": text,
            },
        )
        self.assertEqual(
            d["spans"][14], {"start": 0, "end": 79, "type": "required_word_count", "score": 0.0, "mention": text}
        )
        self.assertEqual(
            d["spans"][15],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_lines_starting_with_bullet_point",
                "score": 0.0,
                "mention": text,
            },
        )
        self.assertEqual(
            d["spans"][16],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_lines_ending_with_ellipsis",
                "score": 0.0,
                "mention": text,
            },
        )
        self.assertEqual(
            d["spans"][17],
            {"start": 0, "end": 79, "type": "fraction_of_duplicate_lines", "score": 0.0, "mention": text},
        )
        self.assertEqual(
            d["spans"][18],
            {
                "start": 0,
                "end": 79,
                "type": "fraction_of_characters_in_duplicate_lines",
                "score": 0.0,
                "mention": text,
            },
        )

    def test_word_count_is_whitespace_sep(self):
        tagger = GopherTagger()
        text = "T h i s \n    \n\n\n    isoneword !!!"
        doc = Document(source="", version="", id="", text=text)
        doc_result = tagger.predict(doc=doc)
        d = doc_result.to_json()
        self.assertEqual(d["spans"][6]["type"], "word_count")
        self.assertEqual(d["spans"][6]["score"], 6.0)

    def test_required_word_count(self):
        tagger = GopherTagger()
        text = "The.and.that"
        doc = Document(source="", version="", id="", text=text)
        doc_result = tagger.predict(doc=doc)
        d = doc_result.to_json()
        self.assertEqual(d["spans"][5]["type"], "required_word_count")
        self.assertEqual(d["spans"][5]["score"], 0.0)

        text = "The and that"
        doc = Document(source="", version="", id="", text=text)
        doc_result = tagger.predict(doc=doc)
        d = doc_result.to_json()
        self.assertEqual(d["spans"][7]["type"], "required_word_count")
        self.assertEqual(d["spans"][7]["score"], 2.0)



import json
import shutil
from contextlib import ExitStack
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from unittest import TestCase

from dolma.cli.__main__ import main

from .utils import (
    clean_test_data,
    download_s3_prefix,
    get_test_prefix,
    load_jsonl,
    upload_s3_prefix,
)

DEDUPE_BY_URL = Path(__file__).parent.parent / "config/dedupe-by-url.json"
DEDUPE_PARAGRAPHS = Path(__file__).parent.parent / "config/dedupe-paragraphs.json"


class TestDeduper(TestCase):
    def setUp(self) -> None:
        self.stack = ExitStack()
        self.test_prefix = get_test_prefix()
        self.local_temp_dir = self.stack.enter_context(TemporaryDirectory()).rstrip("/")

        # upload test data
        upload_s3_prefix(s3_prefix=f"{self.test_prefix}", local_prefix="tests/data/provided/documents/*.gz")

        # copy provided config files to local temp dir
        shutil.copytree(
            "tests/data/provided/documents",
            f"{self.local_temp_dir}/tests/data/provided/documents",
            dirs_exist_ok=True,
        )

    def tearDown(self) -> None:
        clean_test_data(self.test_prefix)
        self.stack.close()

    def test_dedupe_by_url(self):
        with open(DEDUPE_BY_URL, "r") as f:
            config = json.load(f)

        config["documents"][0] = f'{self.local_temp_dir}/{config["documents"][0]}'
        config["bloom_filter"]["file"] = f'{self.local_temp_dir}/{config["bloom_filter"]["file"]}'

        with NamedTemporaryFile("w") as f:
            json.dump(config, f)
            f.flush()

            main(argv=["-c", f.name, "dedupe"])

        expected = load_jsonl("tests/data/expected/dedupe-by-url.json.gz")
        computed = load_jsonl(f"{self.local_temp_dir}/tests/data/provided/attributes/dedupe_by_url/000.json.gz")
        self.assertEqual(expected, computed)

    def test_dedupe_paragraphs(self):
        with open(DEDUPE_PARAGRAPHS, "r") as f:
            config = json.load(f)

        config["documents"][0] = f'{self.local_temp_dir}/{config["documents"][0]}'
        config["bloom_filter"]["file"] = f'{self.local_temp_dir}/{config["bloom_filter"]["file"]}'

        with NamedTemporaryFile("w") as f:
            json.dump(config, f)
            f.flush()

            main(argv=["-c", f.name, "dedupe"])

        expected = load_jsonl("tests/data/expected/dedupe-paragraphs.json.gz")
        computed = load_jsonl(
            f"{self.local_temp_dir}/tests/data/provided/attributes/dedupe_paragraphs/000.json.gz"
        )
        self.assertEqual(expected, computed)

    def test_dedupe_by_url_remote_input(self):
        with open(DEDUPE_BY_URL, "r") as f:
            config = json.load(f)

        config["documents"][0] = f'{self.test_prefix}/{config["documents"][0]}'
        config["bloom_filter"]["file"] = f'{self.local_temp_dir}/{config["bloom_filter"]["file"]}'

        with NamedTemporaryFile("w") as f:
            json.dump(config, f)
            f.flush()

            main(argv=["-c", f.name, "dedupe"])

        download_s3_prefix(self.test_prefix, self.local_temp_dir)

        expected = load_jsonl("tests/data/expected/dedupe-by-url.json.gz")
        computed = load_jsonl(f"{self.local_temp_dir}/tests/data/provided/attributes/dedupe_by_url/000.json.gz")
        self.assertEqual(expected, computed)



import glob
import re
from functools import partial
from itertools import chain
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Tuple, Union
from urllib.parse import urlparse

from fsspec import AbstractFileSystem, get_filesystem_class

__all__ = [
    "glob_path",
    "sub_prefix",
    "add_suffix",
    "sub_suffix",
    "make_relative",
    "mkdir_p",
    "split_path",
    "join_path",
    "is_glob",
    "split_glob",
]


FS_KWARGS: Dict[str, Dict[str, Any]] = {
    "": {"auto_mkdir": True},
}


def _get_fs(path: Union[Path, str]) -> AbstractFileSystem:
    """
    Get the filesystem class for a given path.
    """
    path = str(path)
    protocol = urlparse(path).scheme
    fs = get_filesystem_class(protocol)(**FS_KWARGS.get(protocol, {}))

    # patch glob method to support recursive globbing
    if protocol == "":
        fs.glob = partial(glob.glob, recursive=True)

    return fs


def _escape_glob(s: Union[str, Path]) -> str:
    """
    Escape glob characters in a string.
    """
    r"(?<!\\)[*?[\]]"
    s = str(s)
    s = re.sub(r"(?<!\\)\*", "\u2581", s)
    s = re.sub(r"(?<!\\)\?", "\u2582", s)
    s = re.sub(r"(?<!\\)\[", "\u2583", s)
    s = re.sub(r"(?<!\\)\]", "\u2584", s)
    return s


def _unescape_glob(s: Union[str, Path]) -> str:
    """
    Unescape glob characters in a string.
    """
    s = str(s)
    s = re.sub("\u2581", "*", s)
    s = re.sub("\u2582", "?", s)
    s = re.sub("\u2583", "[", s)
    s = re.sub("\u2584", "]", s)
    return s


def _pathify(path: Union[Path, str]) -> Tuple[str, Path]:
    """
    Return the protocol and path of a given path.
    """
    path = _escape_glob(str(path))
    parsed = urlparse(path)
    path = Path(f"{parsed.netloc}/{parsed.path}") if parsed.netloc else Path(parsed.path)
    return parsed.scheme, path


def _unpathify(protocol: str, path: Path) -> str:
    """
    Return a path from its protocol and path components.
    """
    path_str = _unescape_glob(str(path))
    if protocol:
        path_str = f"{protocol}://{path_str.lstrip('/')}"
    return path_str


def split_path(path: str) -> Tuple[str, Tuple[str, ...]]:
    """
    Split a path into its protocol and path components.
    """
    protocol, _path = _pathify(path)
    return protocol, tuple(_unescape_glob(p) for p in _path.parts)


def join_path(protocol: str, *parts: Union[str, Iterable[str]]) -> str:
    """
    Join a path from its protocol and path components.
    """
    all_parts = (_escape_glob(p) for p in chain.from_iterable([p] if isinstance(p, str) else p for p in parts))
    path = str(Path(*all_parts)).rstrip("/")
    if protocol:
        path = f"{protocol}://{path.lstrip('/')}"
    return _unescape_glob(path)


def glob_path(path: Union[Path, str], hidden_files: bool = False) -> Iterator[str]:
    """
    Expand a glob path into a list of paths.
    """
    path = str(path)
    protocol = urlparse(path).scheme
    fs = _get_fs(path)

    for gl in fs.glob(path):
        gl = str(gl)

        if not hidden_files and Path(gl).name.startswith("."):
            continue

        yield join_path(protocol, gl)


def sub_prefix(a: str, b: str) -> str:
    """
    Return the relative path of b from a.
    """
    prot_a, path_a = _pathify(a)
    prot_b, path_b = _pathify(b)

    if prot_a != prot_b:
        raise ValueError(f"Protocols of {a} and {b} do not match")

    try:
        diff = str(path_a.relative_to(path_b))
    except ValueError:
        diff = join_path(prot_a, path_a.parts)

    return _unescape_glob(diff)


def sub_suffix(a: str, b: str) -> str:
    """
    Remove b from the end of a.
    """
    prot_a, path_a = _pathify(a)
    prot_b, path_b = _pathify(b)

    if prot_b:
        raise ValueError(f"{b} is not a relative path")

    sub_path = re.sub(f"{path_b}$", "", str(path_a))
    sub_prot = f"{prot_a}://" if prot_a else ""

    # need to trim '/' from the end if (a) '/' is not the only symbol in the path or
    # (b) there is a protocol so absolute paths don't make sense
    if sub_path != "/" or sub_prot:
        sub_path = sub_path.rstrip("/")

    return _unescape_glob(sub_prot + sub_path)


def add_suffix(a: str, b: str) -> str:
    """
    Return the the path of a joined with b.
    """
    prot_a, path_a = _pathify(a)
    prot_b, path_b = _pathify(b)

    if prot_b:
        raise ValueError(f"{b} is not a relative path")

    return join_path(prot_a, str(path_a / path_b))


def mkdir_p(path: str) -> None:
    """
    Create a directory if it does not exist.
    """
    if is_glob(path):
        raise ValueError(f"Cannot create directory with glob pattern: {path}")

    fs = _get_fs(path)
    fs.makedirs(path, exist_ok=True)


def make_relative(paths: List[str]) -> Tuple[str, List[str]]:
    """Find minimum longest root shared among all paths"""
    if len(paths) == 0:
        raise ValueError("Cannot make relative path of empty list")

    common_prot, common_parts = (p := _pathify(paths[0]))[0], p[1].parts

    for path in paths:
        current_prot, current_path = _pathify(path)
        if current_prot != common_prot:
            raise ValueError(f"Protocols of {path} and {paths[0]} do not match")

        current_parts = current_path.parts
        for i in range(min(len(common_parts), len(current_parts))):
            if common_parts[i] != current_parts[i]:
                common_parts = common_parts[:i]
                break

    if len(common_parts) > 0:
        common_path = (f"{common_prot}://" if common_prot else "") + str(Path(*common_parts))
        relative_paths = [sub_prefix(path, common_path) for path in paths]
    else:
        common_path = f"{common_prot}://" if common_prot else ""
        relative_paths = [_unpathify("", _pathify(path)[1]) for path in paths]

    return common_path, relative_paths


def is_glob(path: str) -> bool:
    """
    Check if a path contains a glob wildcard.
    """
    return bool(re.search(r"(?<!\\)[*?[\]]", path))


def split_glob(path: str) -> Tuple[str, str]:
    """
    Partition a path on the first wildcard.
    """
    if not is_glob(path):
        return path, ""

    protocol, parts = split_path(path)

    i = min(i for i, c in enumerate(parts) if is_glob(c))

    path = join_path(protocol, *parts[:i])
    rest = join_path("", *parts[i:])
    return path, rest



import logging
import multiprocessing
import tempfile
from contextlib import ExitStack
from queue import Queue
from typing import Dict, List, Union

import msgspec
import smart_open

from .data_types import InputSpec, OutputSpec
from .errors import DolmaFatalError, DolmaRetryableFailure, DolmaShardError
from .parallel import BaseParallelProcessor
from .paths import join_path, make_relative, split_glob, split_path
from .registry import TaggerRegistry
from .utils import make_variable_name


def _make_paths_from_substitution(paths: List[str], find: str, replace: str) -> List[str]:
    """
    Utility function to make paths using a find/replace substitution. This is useful if you want to
    create a destination path from a source path by replacing part of the source path with something else.

    For example, if you have a source path of `current_paths = ["s3://bucket/data/documents/**.json.gz"]` and
    you want to replace `documents` with `attributes`, you can use this function to do that. by calling
    `_make_paths_from_substitution(current_paths, "documents", "attribute")`. This will return the following
    list `["s3://bucket/data/attributes"]`. Note how glob patterns are removed from the paths.
    """
    new_paths: List[str] = []
    for curr in paths:
        curr_pre_glob, _ = split_glob(curr)
        curr_prot, curr_parts = split_path(curr_pre_glob)
        find_dir_index = curr_parts.index(find)

        if not curr_pre_glob.strip():
            raise RuntimeError(f"Path '{curr}' contains a wildcard at the beginning. ")
        elif find_dir_index < 0:
            raise RuntimeError(f"Path '{curr}' does not contain a '{find}' component.")

        dst_parts = [p if i != find_dir_index else replace for i, p in enumerate(curr_parts)]
        new_paths.append(join_path(curr_prot, dst_parts))

    return new_paths


def _make_paths_from_prefix(paths: List[str], prefix: str) -> List[str]:
    """
    Utility function to make paths using a prefix. This is useful if you want to create a destination path
    from a source path by prepending a prefix to the source path.

    To create destination paths, we first find the longest common prefix among all source paths. Then, we
    we remove the prefix from each source path and prepend the new prefix to each source path. For example,
    if you have a source path of
    ```
    current_paths = [
        "s3://bucket/data/documentsA/**.json.gz",
        "s3://bucket/data/documentsB/**.json.gz",
    ]
    ```
    and you want to replace `s3://bucket/data/` with `s3://bucket/attributes/`, you can use this function
    to do that. by calling `_make_paths_from_prefix(current_paths, "s3://bucket/attributes/")`. This will
    return the following list

    ```
    [
        "s3://bucket/attributes/documentsA/",
        "s3://bucket/attributes/documentsB/",
    ]
    ```

    Note how glob patterns are removed from the paths.
    """

    new_paths: List[str] = []
    prefix_prot, prefix_path = split_path(prefix)
    _, relative_paths = make_relative(paths)

    for curr_path in relative_paths:
        base_curr_path, _ = split_glob(curr_path)
        new_paths.append(join_path(prefix_prot, prefix_path, base_curr_path))

    return new_paths


class TaggerProcessor(BaseParallelProcessor):
    @classmethod
    def get_logger(cls) -> logging.Logger:
        return logging.getLogger(cls.__name__)

    @classmethod
    def increment_progressbar(  # type: ignore
        cls,
        queue,  # queue must be the first argument, and it should be a positional-only argument
        /,
        files: int = 0,
        documents: int = 0,
    ) -> Dict[str, int]:
        """We override this method to specify which units we want to keep track of in a progress bar.
        Specifically, we keep track of files and documents in this example. Their default value must be zero."""

        # we call the super method to increment the progress bar
        return super().increment_progressbar(queue, files=files, documents=documents)

    @classmethod
    def process_single(
        cls,
        source_path: str,
        destination_path: str,
        queue: "Queue",
        **kwargs,
    ):
        """Lets count run the taggers! We will use the destination path to save each tagger output."""

        # get names of taggers
        taggers_names = kwargs.get("taggers_names", None)
        if taggers_names is None:
            raise RuntimeError("Taggers not in kwargs, this is a bug! Please report it.")
        elif not isinstance(taggers_names, list) or not all(isinstance(t, str) for t in taggers_names):
            raise RuntimeError("Taggers are in the wrong format, this is a bug! Please report it.")
        taggers = {make_variable_name(t): TaggerRegistry.get(t)() for t in taggers_names}

        # get name of experiment
        experiment_name = kwargs.get("experiment_name", None)
        if experiment_name is None:
            raise RuntimeError("Experiment name not in kwargs, this is a bug! Please report it.")
        experiment_name = make_variable_name(experiment_name)

        # skip on failure
        skip_on_failure = kwargs.get("skip_on_failure", False)

        # interval at which to update the progress bar; will double if it gets
        # too full
        update_interval = 1

        # running document count; gets reset every time we update the progress
        # bar
        docs_cnt = 0

        # creating dedicated encoders/decoders speeds up the process
        encoder = msgspec.json.Encoder()
        decoder = msgspec.json.Decoder(InputSpec)

        with ExitStack() as stack:
            try:
                # open each file for reading and writing. We use open_file_for_read to handle s3 paths and
                # download the file locally if needed, while gzip.open is used to read and write gzipped files.
                in_stream = stack.enter_context(smart_open.open(source_path, "rt", encoding="utf-8"))
                out_stream = stack.enter_context(smart_open.open(destination_path, "wt", encoding="utf-8"))

                for raw in in_stream:
                    row = decoder.decode(raw)

                    # running the taggers and merging them flat
                    attributes = {}
                    for tagger_name, tagger in taggers.items():
                        for key_name, key_value in tagger.tag(row).items():
                            key_name = f"{experiment_name}__{tagger_name}__{make_variable_name(key_name)}"
                            attributes[key_name] = key_value

                    # make output file
                    output = OutputSpec(source=row.source, id=row.id, attributes=attributes)

                    # write the output to the output file
                    out_stream.write(encoder.encode(output).decode("utf-8") + "\n")

                    # increment the number of documents processed so far
                    docs_cnt += 1

                    if docs_cnt % update_interval == 0:
                        # update the progress bar every 1000 documents to prevent
                        # buffering
                        cls.increment_progressbar(queue, documents=docs_cnt)
                        docs_cnt = 0

                        if queue.qsize() >= multiprocessing.cpu_count():
                            # double the update interval if the queue is full
                            update_interval *= 2

            except Exception as e:
                # handle any exception that might have occurred
                msg = f"Failed to process {source_path} due to {e.__class__.__name__}: {' '.join(e.args)}"
                if e.__class__.__name__ == "IncompleteReadError":
                    # Intermittent error that occurs when reading from S3
                    raise DolmaRetryableFailure(msg) from e
                else:
                    if skip_on_failure:
                        raise DolmaShardError(msg) from e
                    else:
                        raise DolmaFatalError(msg) from e

        # increment the files progress bar
        cls.increment_progressbar(queue, files=1, documents=docs_cnt)


def create_and_run_tagger(
    documents: List[str],
    experiment: str,
    taggers: List[str],
    destination: Union[None, str, List[str]] = None,
    metadata: Union[None, str, List[str]] = None,
    debug: bool = False,
    seed: int = 0,
    ignore_existing: bool = False,
    skip_on_failure: bool = False,
    retries_on_error: int = 0,
    num_processes: int = 1,
):
    """This function creates a tagger and runs it on a list of documents.

    Args:
        documents (List[str]): List of documents to run the taggers on. Each element of the list is a path to
            a file containing documents in json lines format, or a glob pattern that matches such files.
        experiment (str): The name of the experiment. This will be used to prefix the names of the attributes,
            as well as the name of the directory where the outputs will be saved in `destination`.
        taggers (List[str]): List of taggers to run. Each element of the list is the name of a tagger.
        destination (Union[None, str, List[str]], optional): The path where the outputs will be saved. If
            `None`, the outputs will be saved in a directory parallel to the directory containing the
            documents, with the same name as `experiment`. If a string, paths corresponding to each element
            of `documents` will be created by determining a relative path from the directory containing the
            documents.
        metadata (Union[None, str, List[str]], optional): Location where to save metadata that keeps track of
            which documents have been processed. If `None`, the metadata will be saved in a temporary directory.
        debug (bool, optional): Whether to run in debug mode. Defaults to False.
        seed (int, optional): The seed to use for the random number generator. Defaults to 0.
        ignore_existing (bool, optional): Whether to ignore existing outputs and re-run the taggers.
            Defaults to False.
        skip_on_failure (bool, optional): Whether to skip a document if it fails to process. Defaults to False.
        retries_on_error (int, optional): Number of times to retry processing a document if it fails.
            Defaults to 0 (fail immediately)
        num_processes (int, optional): Number of processes to use. Defaults to 1.
    """

    if destination is None:
        try:
            destination = _make_paths_from_substitution(documents, "documents", f"attributes/{experiment}")
        except Exception as e:
            raise RuntimeError("Could not make destination paths from documents paths") from e
    elif isinstance(destination, str):
        try:
            destination = _make_paths_from_prefix(documents, join_path(destination, experiment))
        except Exception as e:
            raise RuntimeError(f"Could not make destination paths from prefix {destination}") from e

    metadata = metadata or tempfile.mkdtemp()
    if isinstance(metadata, str):
        try:
            metadata = _make_paths_from_prefix(documents, metadata)
        except Exception as e:
            raise RuntimeError(f"Could not make metadata paths from prefix {metadata}") from e

        tagger = TaggerProcessor(
            source_prefix=documents,
            destination_prefix=destination,
            metadata_prefix=metadata,
            debug=debug,
            seed=seed,
            ignore_existing=ignore_existing,
            retries_on_error=retries_on_error,
            num_processes=num_processes,
        )
        tagger(
            experiment_name=experiment,
            taggers_names=taggers,
            skip_on_failure=skip_on_failure,
        )



"""

Builds a dataset for training a FastText model from 2 or more pretraining datasets.

@rauthur

"""

import argparse
import gzip
import json
import os
import random
from dataclasses import dataclass
from functools import partial
from multiprocessing import Manager, Pool, Process, Queue
from threading import Event
from typing import Generator, List, Optional

from smashed.utils.io_utils import (
    open_file_for_read,
    open_file_for_write,
    recursively_list_files,
)

from .data_types import TextSlice
from .ft_tagger import BaseFastTextTagger
from .utils import split_paragraphs, split_sentences

_WRITER_EXIT_MSG = "__WRITE__EXIT__"


@dataclass
class Config:
    target_path: str
    sample_paths: List[str]
    out_path: str
    mode: str
    newlines: str
    n_proc: int
    n_segments: Optional[int]
    pos_label: str
    neg_label: str


def gzip_open(file, mode, **open_kwargs):
    return gzip.open(filename=file, mode=mode, **open_kwargs)


def _split(text: str, config: Config) -> Generator[TextSlice, None, None]:
    if config.mode == BaseFastTextTagger.SENTENCE_LEVEL_TAGGER:
        for sentence in split_sentences(text):
            yield sentence

    elif config.mode == BaseFastTextTagger.PARAGRAPH_LEVEL_TAGGER:
        for paragraph in split_paragraphs(text):
            yield paragraph

    elif config.mode == BaseFastTextTagger.DOCUMENT_LEVEL_TAGGER:
        yield TextSlice(doc=text, start=0, end=len(text))

    else:
        raise RuntimeError(f"Unknown data split mode: {config.mode}")


@dataclass
class ReadResult:
    examples: List[str]


def process_file(config: Config, q: "Queue[str]", flag: Event, label: str, fn):
    # Check a global exit flag and stop processing file
    if flag.is_set():
        return

    print(f"Processing {fn}")

    with open_file_for_read(fn, "rb", open_fn=gzip_open) as f:
        for line in f:
            # Abort part way through processing this file is flag set
            if flag.is_set():
                return

            # Expected JSONL format following OLMo data spec
            data = json.loads(line.decode("utf-8"))
            line_text = data["text"]

            if len(line_text) == 0:
                continue

            for slice in _split(line_text, config):
                final_text = slice.text

                if "\n" in final_text:
                    if config.newlines == "replace":
                        final_text = final_text.replace("\n", " ")
                    elif config.newlines == "skip":
                        continue

                q.put(f"__label__{label} {final_text}")


def write_results(config: Config, q: "Queue[str]", flag: Event):
    written = 0

    with open_file_for_write(config.out_path) as o:
        while True:
            msg = q.get()

            if msg == _WRITER_EXIT_MSG:
                break

            if not flag.is_set():
                o.write(q.get())
                o.write("\n")
                written += 1

            if config.n_segments is not None and written >= config.n_segments:
                flag.set()
                written = 0


def process_paths(paths: List[str], config: Config, q: "Queue[str]", flag: Event, label: str):
    fns = [fn for p in paths for fn in recursively_list_files(p)]
    random.shuffle(fns)

    work_fn = partial(process_file, config, q, flag, label)

    with Pool(processes=max(1, config.n_proc - 1)) as pool:
        pool.map(work_fn, fns)
        pool.close()
        pool.join()


def main(config: Config):
    random.seed(117)

    with Manager() as manager:
        q: "Queue[str]" = manager.Queue()  # type: ignore
        flag = manager.Event()

        writer = Process(target=write_results, args=(config, q, flag))
        writer.start()

        # Generate expected number of positive examples
        process_paths([config.target_path], config, q, flag, config.pos_label)

        # Reset early exit flag as all positive examples processed
        flag.clear()

        # Generate expected number of negative examples
        process_paths(config.sample_paths, config, q, flag, config.neg_label)

        q.put(_WRITER_EXIT_MSG)
        writer.join()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-t",
        "--target",
        required=True,
        type=str,
        help="Local or remote path including OLMo formatted TARGET dataset",
    )
    parser.add_argument(
        "-s",
        "--sample",
        required=True,
        type=str,
        nargs="+",
        help="Sample these paths to create negative examples",
    )
    parser.add_argument(
        "-p",
        "--processes",
        type=int,
        required=False,
        default=os.cpu_count(),
        help="Number of processes to launch",
    )
    parser.add_argument(
        "-m",
        "--mode",
        type=str,
        required=True,
        choices=[
            BaseFastTextTagger.SENTENCE_LEVEL_TAGGER,
            BaseFastTextTagger.PARAGRAPH_LEVEL_TAGGER,
            BaseFastTextTagger.DOCUMENT_LEVEL_TAGGER,
        ],
        help="Output examples at this level",
    )
    parser.add_argument(
        "-o",
        "--output-filename",
        type=str,
        required=True,
        help="Path to write the processed result (can be on S3)",
    )
    parser.add_argument(
        "--n-segments",
        type=int,
        required=False,
        help="Stop after generating this many segments (e.g., sentences)",
    )
    parser.add_argument(
        "--newlines",
        type=str,
        required=False,
        choices=["skip", "keep", "replace"],
        default="skip",
        help="Skip, keep or replace with ' ' examples with newlines after splitting",
    )
    parser.add_argument(
        "--pos-label",
        type=str,
        required=False,
        default="pos",
        help="Use this class label for positive instances",
    )
    parser.add_argument(
        "--neg-label",
        type=str,
        required=False,
        default="neg",
        help="Use this class label for negative instances",
    )

    args = parser.parse_args()
    config = Config(
        target_path=args.target,
        sample_paths=args.sample,
        out_path=args.output_filename,
        mode=args.mode,
        newlines=args.newlines,
        n_proc=args.processes,
        n_segments=args.n_segments,
        pos_label=args.pos_label,
        neg_label=args.neg_label,
    )

    main(config)



from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from omegaconf import OmegaConf as om

from dolma import mixer
from dolma.cli import BaseCli, field, print_config
from dolma.cli.shared import WorkDirConfig


@dataclass
class StreamOutputConfig:
    path: str = field(help="Path where to write the mixed documents to. Required.")
    max_size_in_bytes: int = field(
        default=2 * 2**30, help="Maximum size of the output file in bytes. Defaults to 2GB."
    )
    discard_fields: List[str] = field(default=[], help="List of fields to discard from the output documents.")


@dataclass
class FilterConfig:
    include: List[str] = field(default=[], help="JSONPath expressions to include documents")
    exclude: List[str] = field(default=[], help="JSONPath expressions to exclude documents")


@dataclass
class SpanReplacementConfig:
    span: str = field(help="JSONPath expression for the span to replace")
    min_score: float = field(default=0.5, help="Minimum score for the span to be replaced")
    replacement: str = field(default="", help="Replacement for the span")


@dataclass
class StreamConfig:
    name: str = field(help="Name of the stream. Required.")
    documents: List[str] = field(default=[], help="Paths to the documents to be mixed. Required.")
    output: StreamOutputConfig = field(
        default=StreamOutputConfig(), help="Configuration for the output of the stream."
    )
    attributes: List[str] = field(default=[], help="List of attributes files to used for mixing.")
    filter: Optional[FilterConfig] = field(  # pyright: ignore
        default=None, help="Configuration for filtering documents."
    )
    span_replacement: List[SpanReplacementConfig] = field(default=[], help="Configuration for replacing spans.")


@dataclass
class MixerConfig:
    streams: List[StreamConfig] = field(default=[], help="List configurations of streams to be mixed")
    work_dir: WorkDirConfig = field(default=WorkDirConfig(), help="Configuration for temporary work directories.")
    processes: int = field(default=1, help="Number of processes to use for mixing. By default 1 process is used.")


class MixerCli(BaseCli):
    CONFIG = MixerConfig

    @classmethod
    def run(cls, parsed_config: MixerConfig):
        dict_config: Dict[str, Any] = {
            "work_dir": {"input": parsed_config.work_dir.input, "output": parsed_config.work_dir.output},
            "processes": parsed_config.processes,
            "streams": [],
        }

        for stream_config in parsed_config.streams:
            stream_config_dict: Dict[str, Any] = {}

            if stream_config.filter is not None:
                stream_config_dict["filter"] = {}

                if len(stream_config.filter.include):
                    stream_config_dict["filter"]["include"] = list(stream_config.filter.include)

                if len(stream_config.filter.exclude):
                    stream_config_dict["filter"]["exclude"] = list(stream_config.filter.exclude)

                if len(stream_config_dict["filter"]) == 0:
                    raise ValueError("Either `include` or `exclude` must be specified for filter")

            for span_replacement in stream_config.span_replacement:
                stream_config_dict.setdefault("span_replacement", []).append(
                    {
                        "span": span_replacement.span,
                        "min_score": span_replacement.min_score,
                        "replacement": span_replacement.replacement,
                    }
                )

            if "span_replacement" not in stream_config_dict and "filter" not in stream_config_dict:
                raise ValueError("Either `filter` or `span_replacement` must be specified")

            stream_config_dict["name"] = stream_config.name
            stream_config_dict["documents"] = list(stream_config.documents)
            stream_config_dict["attributes"] = list(stream_config.attributes)
            stream_config_dict["output"] = {
                "path": stream_config.output.path,
                "max_size_in_bytes": stream_config.output.max_size_in_bytes,
                "discard_fields": om.to_container(stream_config.output.discard_fields),
            }

            if len(stream_config_dict["documents"]) == 0:
                raise ValueError("No documents to mix")

            dict_config["streams"].append(stream_config_dict)

        if len(dict_config["streams"]) == 0:
            raise ValueError("No streams to mix")

        print_config(dict_config)
        return mixer(dict_config)
